#' import.preprocess.data.int
#'
#' Internal function
#'
#' @param gauge string of entire file path of gauge .AM file
#' @param metadata metadata file containing gauge numbers and names
#'
#' @return Pre-processed data is returned.
#' @export
#' @keywords internal
import.preprocess.data.int <- function(gauge, metadata){
  # Import and set up data for analysis
  # Read all the lines of the file - this gives a warning if some lines are not complete
  # Re-write the warning to be more informative
  warningMessage <- paste0("Warning: Incomplete line found in ", gauge, ". This is likely to be a missing stage value and therefore not a problem but it is worth checking that no flow data are missing in any years. Any years with missing flow data have been removed so that the distribution fitting will work.")
  tryCatch(header <- readLines(gauge), warning=function(w) print(warningMessage))
  header <- suppressWarnings(readLines(gauge))
  gaugeNumber <- header[grep("\\[STATION NUMBER\\]", header, ignore.case = T) + 1]  %>% str_replace("\t\t","")  # should always be the second row but just to make sure
  rejectedLine <- grep("^\\[AM Rejected\\]", header, ignore.case = T)  # find the rejected years - no cases where these are in tab delimited files so not accounted for
  skip <- grep("^\\[AM Values\\]", header, ignore.case = T)  # find which row the line "[AM Values]" is on (line before data) so we can skip that number of rows
  # Find rejected years
  if (length(rejectedLine) > 0){
    rejected <- header[(rejectedLine+1):(skip-2)]
    if(!is.na(word(rejected[1], 2, sep=","))){  # only need first element to check if there is only one value given (which would result in this option failing)
      rejectedEval <- paste0("c(", paste("seq(", as.numeric(word(rejected, 1, sep = ",")), ",", as.numeric(word(rejected, 2, sep = ",")), ")", sep = "", collapse = ","), ")")
      rejectedYears <- unique(eval(parse(text = rejectedEval)))
      } else{
      rejectedYears <- word(rejected, 1, sep = ",")
    }
    rejectedYearsString <- paste(rejectedYears, collapse = " & ")
  } else{
    rejected <- ""
    rejectedYears <- ""
    rejectedYearsString <- ""
  }

  gaugeRiver <- metadata$RIVER[metadata$STATION == gaugeNumber]
  gaugeLocation <- metadata$LOCATION[metadata$STATION == gaugeNumber]
  gaugeNameComb <- ifelse(gaugeNumber %in% metadata$STATION, paste0(gaugeRiver, " at ", gaugeLocation), "")
  gaugeName <- ifelse(grepl("/", gaugeNameComb), str_replace(gaugeNameComb, "/", "-"), gaugeNameComb)

  inData = fread(gauge, header = FALSE, skip = skip, nrows = length(header) - skip - 1, stringsAsFactors = FALSE)
  inData = data.frame(inData)
  names(inData)[1:2] <- c("Date", "Flow")  # set column headers for first two columns
  inData$Year <- year(as.Date(inData$Date, format = "%d %b %Y"))  # extract year from flow date (temporary column for determining water year)
  inData$Month <- month(as.Date(inData$Date, format = "%d %b %Y"))  # extract month from flow date (temporary column for determining water year)
  inData$WaterYear <- ifelse(inData$Month < 10, inData$Year - 1, inData$Year)  # determine the water year for each event
  if (NA %in% inData$WaterYear){
    date_form <- "%d-%b-%y"
    inData$Year <- year(as.Date(inData$Date, format = date_form))  # extract year from flow date (temporary column for determining water year)
    inData$Month <- month(as.Date(inData$Date, format = date_form))  # extract month from flow date (temporary column for determining water year)
    inData$WaterYear <- ifelse(inData$Month < 10, inData$Year - 1, inData$Year)  # determine the water year for each event
  }
  inData$GaugeNumber <- gaugeNumber  # add a column containing the gauge number
  inData$GaugeName <- gaugeName  # add a column containing the gauge name
  inData$FirstWaterYear <- min(inData$WaterYear)  # add a column containing the first water year in the record
  inData$LastWaterYear <- max(inData$WaterYear)  # add a column containing the last water year in the record
  # Get sequence of years from first water year in record to last
  yearRange <- seq(min(inData$WaterYear), max(inData$WaterYear), by = 1)
  # Select the years which are not in the record
  missingYears <- yearRange[!yearRange %in% inData$WaterYear]
  # Add missing and rejected years as columns
  inData$Missing <- ifelse(length(missingYears > 0), paste(missingYears, collapse = " & "), "")
  inData$Rejected <- rejectedYearsString

  return(list(inData = inData, rejectedYears = rejectedYears, gaugeNumber = gaugeNumber))
}


#' import.preprocess.data
#'
#' Function to pre-process data to put it into the appropriate format for analysis.
#' Mainly used internally, but can be used by user, for example, to plot results for multiple gauges at once.
#'
#' @param gauge string of entire file path of gauge .AM file
#' @param metadata metadata file containing gauge numbers and names
#'
#' @return Pre-processed data is returned as well as any warnings or errors associated with the pre-processing.
#' @export
import.preprocess.data <- function(gauge, metadata){
  warn <- err <- NULL
  res <- withCallingHandlers(
    tryCatch(import.preprocess.data.int(gauge, metadata), error=function(e){
                    err <<- conditionMessage(e)
                    NULL
                  }), warning=function(w){
                    warn <<- append(warn, conditionMessage(w))
                  })
  list(res=res, warn=warn, err=err)
}


#' fitWarning
#'
#' Internal function to fit distribution and catch errors and warnings.
#'
#' @param Flow Either a numeric vector or the name of a variable in all_data to which the distribution is to be fitted.
#' @param all_data A data frame containing response variable (Flow) and any covariates.
#' @param distFamily An object of class 'texmexFamily'. Can be gev or glo.
#' @param modelsList List of length two containing the regression formula for the location and scale parameters.
#' @param methodType String of method to be used to fit distribution.
#'
#' @return Returns fitted distribution (res) as well as warning messages (warn) and error messages (err) from fitting, if produced.
#' @export
#' @keywords internal
fitWarning <- function(Flow, all_data, distFamily, modelsList, methodType = "optim"){
  warn <- err <- NULL
  res <- withCallingHandlers(
    tryCatch(evm(y = Flow, data = all_data, family = eval(parse(text = distFamily)), mu = eval(parse(text=modelsList[1])),
                 phi = eval(parse(text=modelsList[2])), method = methodType, verbose = FALSE, cores = 2), error=function(e){
                   err <<- conditionMessage(e)
                   NULL
                 }), warning=function(w){
                   warn <<- append(warn, conditionMessage(w))
                 })
  return(list(res=res, warn=warn, err=err))
}

#' p.value.comparison
#'
#' Internal function to compute a p-value comparing fit of two models using a likelihood ratio test.
#' @keywords internal
p.value.comparison <- function(model1, model2) {
  chiSqStatistic <- 2*(-model1$map$loglik-(-model2$map$loglik))
  diffPar <- length(model2$map$par) - length(model1$map$par)
  pValue <- 1-pchisq(chiSqStatistic, diffPar)
  return(pValue)
}


#' likelihood.ratio.testing
#'
#' Internal function to carry out likelihood ratio testing to find best fit based
#' on four distributions (time covariate only). Uses forward selection to compare
#' best fits based on negative log-likelihood.
#' @keywords internal
likelihood.ratio.testing <- function(statFit, varLocFit, varScaleFit, varLocScaleFit){

  candModels <- list(varLocFit, varScaleFit)  # set the candidate models to be those with one additional parameter to the stationary model (null model)

  pValCandModels <- sapply(candModels, p.value.comparison, model1 = statFit)  # extract p-values from comparing with the stationary model

  # If the lower of the two p-values is less than 0.05 then select the information for that model and set as the best fit
  if(min(pValCandModels) < 0.05) {

    indx <- which.min(pValCandModels)  # find which of the values is the minimum

    bestModelLabel <- ifelse(indx==1, "Non-stationary (varying location)", "Non-stationary (varying scale)")  # set label for which is the minimum

    bestModel <- candModels[[indx]]  # select the best model as that with the minimum p-value
    p <- pValCandModels[indx]  # extract the p value for the best model

    # Compare the best model with the varying location and scale option
    pValBoth <- p.value.comparison(bestModel, varLocScaleFit)

    # If the p-value is less than pValBoth, then select the varying location and scale option as the best fit model and compare with stationary model
    if (pValBoth < 0.05) {
      bestModel <- varLocScaleFit
      bestModelLabel <- "Non-stationary (varying location and scale)"
      p <- p.value.comparison(statFit, bestModel)
    }
  } else {  # else set the stationary model as the best one
    bestModel <- statFit
    bestModelLabel <- "Stationary"
    p <- min(pValCandModels)
  }

  return(bestModelLabel)

}


#' plot.diagnostic
#'
#' Internal function used in plot.diagnostic.warn.err. Plots diagnostic plots for a given model (stationary or non-stationary).
#'
#' @param dist distribution type used to fit model
#' @param model fitted model
#' @param outPath file path for writing out plots
#' @param gaugeNumber gauge number
#' @param gaugeName gauge name
#' @param gaugeData gaugeData used in model fitting
#' @param pp.plotTitle optional plot title for P-P plot
#' @param qq.plotTitle optional plot title for Q-Q plot
#'
#' @return Returns diagnostic plots which are written to a .png file.
#' @keywords internal
plot.diagnostic <- function(dist, model, outPath, gaugeNumber, gaugeName, gaugeData, pp.plotTitle = NULL, qq.plotTitle = NULL){

  gg_theme <- theme_bw() + theme(title = element_text(size=15),
                                 axis.title = element_text(size=10),
                                 axis.text = element_text(size=10),
                                 strip.text = element_text(size=10),
                                 plot.title = element_text(hjust=0.5))

  covs <- model$map$formulae
  loc <- Reduce(paste, deparse(covs$mu))  # convert formula to string
  scale <- Reduce(paste, deparse(covs$phi))  # convert formula to string
  fitType <- ifelse(loc == "~1" & scale == "~1", "Stationary", "Non-stationary")
  covsInfo <- paste0("Location: ", as.character(loc), "; Scale: ", as.character(scale))
  main <- ifelse(nchar(gaugeName) > 0, paste0("Gauging station ", gaugeNumber, " (", gaugeName, ") (", dist, ") \n ",
                                              fitType, " (", covsInfo, ")"),
                 paste0("Gauging station ", gaugeNumber, " (", dist, ") \n ", fitType, " (", covsInfo, ")"))
  if(is.null(pp.plotTitle)) pp.plotTitle <- "Residual P-P plot"
  if(is.null(qq.plotTitle)) qq.plotTitle <- "Residual Q-Q plot (Gumbel scale)"
  d.plotTitle <- "Observed data"
  pp.xlab <- qq.xlab <- "Model"
  pp.ylab <- qq.ylab <- "Empirical"

  # Model setup
  n <- length(model$map$data$y)

  # Parameters
  params <- predict(model$map, ci.fit=FALSE, type="lp", unique.=FALSE)

  loc <- params$obj$link[,1]
  scale <- exp(params$obj$link[,2])
  shape <- params$obj$link[,3]

  # P-P plot
  if(dist == "GEV"){
    dist.pp <- sapply(1:n, function(i) texmex::pgev(model$map$data$y[i], mu=loc[i], sigma=scale[i], xi=shape[i]))
  } else if(dist == "GLO"){
    dist.pp <- sapply(1:n, function(i) texmex::pglo(model$map$data$y[i], mu=loc[i], sigma=scale[i], xi=shape[i]))
  }

  probs <- (1:n)/(n+1)
  pp <- data.frame(model = probs, empirical = sort(dist.pp))

  g <- ggplot(aes(x=model, y=empirical), data=pp) + geom_point() + geom_abline(intercept=0, slope=1) + gg_theme + ggtitle(pp.plotTitle) + xlab(pp.xlab) + ylab(pp.ylab)

  # Q-Q plot
  # "Gumbel.data" is data after transformation to standard Gumbel margins.
  Gumbel.data <- -log(-log(dist.pp))
  Gumbel.quantile <- -log(-log(probs))

  qq <- data.frame(model = sort(Gumbel.quantile), empirical = sort(Gumbel.data))

  h <- ggplot(aes(model, empirical), data=qq) + geom_point() + geom_abline(intercept=0, slope=1) + gg_theme + ggtitle(qq.plotTitle) + xlab(qq.xlab) + ylab(qq.ylab)

  d <- ggplot(gaugeData, aes(WaterYear, Flow)) + geom_point() + gg_theme + ggtitle(d.plotTitle) + xlab("Water year") + ylab(expression(paste("Flow (", m^3, "/s)")))

  dgh <- grid.arrange(d, g, h, nrow = 1, top = textGrob(main, gp = gpar(fontface = "bold", fontsize = 20)))

  ggsave(outPath, dgh, width = 15, height = 7, dpi = 120)

  dev.off()

  return(list(data=d, pp=g, qq=h))
}


#' plot.diagnostic.warn.err
#'
#' Internal function to plot diagnostic plots and catch any warnings and errors so they can be written to a file.
#' Inputs are the same as above.
#' @keywords internal
plot.diagnostic.warn.err <- function(dist, model, outPath, gaugeNumber, gaugeName, gaugeData){
  warn <- err <- NULL
  res <- withCallingHandlers(
    tryCatch(plot.diagnostic(dist, model, outPath, gaugeNumber, gaugeName, gaugeData), error=function(e){
      err <<- conditionMessage(e)
      NULL
    }), warning=function(w){
      warn <<- append(warn, conditionMessage(w))
    })
  list(warn=warn, err=err, plots=res)
}


#' detrend.fn
#'
#' Internal function to detrend covariate data based on a selected covariate.
#'
#' @param nondt_data original non-detrended data
#' @param detrend.cov string of covariate to use for detrending (usually time, i.e. "ScaledWaterYearIndex")
#' @param regressors list of regressors
#'
#' @return Returns detrended data.
#' @keywords internal
detrend.fn <- function(nondt_data, detrend.cov, regressors){
  # If length of detrend.cov is not equal to 1 or detrend.cov is not in regressors then produce an error message
  if(length(detrend.cov) != 1 | !(detrend.cov %in% regressors) ) {
    stop("Detrending must be done based on one variable that must be included in the input data")
  }

  trend_var <- detrend.cov
  predict_var <- setdiff(regressors, trend_var)

  # Take detrended covariates as new regressors
  dt_data <- nondt_data
  dt_data[, predict_var] <- sapply(nondt_data[, predict_var], function(x) lm(x ~ nondt_data[, trend_var])$residuals)

  return(dt_data)  # return detrended data
}
