#' BestModel
#'
#' Internal function. Fits models using a range of covariate combinations and selects best-fitting
#' model among all possible candidate models using AIC or BIC.
#'
#' @param dist distribution to use for fitting ("GEV" or "GLO")
#' @param all_data_raw data.frame of response variable (1st column) and regressors (contains raw water years and WaterYearIndex for output files and plots)
#' @param detrend.cov optional character string. If included, instructs to detrend other covariates based on covariate in that string (usually "ScaledWaterYearIndex", representing time).
#' @param option optional - must be 1, 2 or 3: 1 - fit model to all possible combinations of covariates; 2 - fit model to combinations which only use one covariate and/or time; 3 - fit model to combinations which use one covariate at a time in the location and/or scale parameter.
#' @param best_fit_method optional string - method of automatically selecting a best fit model. Must be "AIC" or "BIC".
#'
#' @return The function outputs the following:
#' @return best_model - details of best fitting model selected based on AIC or BIC.
#' @return best_model_formula - details of the formula used to fit best_model.
#' @return stat_model - details of stationary model.
#' @return all_models_metadata - a metadata file associated with fitting models to all combinations of covariates selected by the user. Contains location and scale parameter covariate regression equations, warnings and errors associated with the fitting, parameters from fitting, AIC and BIC of fitted models and the rank of each model based on AIC and BIC.
#' @keywords internal
BestModel <- function(dist, all_data_raw, detrend.cov, option, best_fit_method) {

  all_data <- all_data_raw[, !names(all_data_raw) %in% c("WaterYear", "WaterYearIndex")]

  # Check that option equals 1, 2 or 3 and stop and print warning if not
  if(!(option == 1 | option == 2 | option == 3)) stop("option must be 1, 2 or 3", call. = FALSE)

  # Check that best_fit_method equals "AIC" or "BIC" and stop and print warning if not
  if(!(best_fit_method == "AIC" | best_fit_method == "BIC")) stop("best_fit_method must be \"AIC\" or \"BIC\"", call. = FALSE)

  # Extract predictor variables - keep all covariates and exclude 'Flow' column
  covariateData <- all_data[, !names(all_data) %in% "Flow"]
  regressors <- names(covariateData)

  # Detrend covariates based on what has been specified in detrend.cov - should be time-based covariate
  # If detrend.cov is not null (i.e. user has inputted something)
  if(!is.null(detrend.cov)) {
    all_data <- detrend.fn(all_data, detrend.cov, regressors)
  }

  # Reset covariateData now that the data have been detrended
  covariateData <- all_data[, !names(all_data) %in% "Flow"]

  #############################################################################################################################
  # Matrix of different combinations of regressors to be included in models (the first half of the columns is for the location
  # parameter and the second half is for the scale parameter)

  ################################## OPTION 1: Find every possible combination of covariates ##################################
  if(option==1){
    regMat <- expand.grid(rep(list(c(TRUE,FALSE)), length(regressors)*2))  # each regressor can be TRUE or FALSE
    names(regMat) <- paste(rep(regressors, 2))
  }

  ######## OPTION 2: Remove rows so we are only using one covariate and/or time rather than lots of covariates at once ########
  if(option==2){
    # As above
    regMat <- expand.grid(rep(list(c(TRUE,FALSE)), length(regressors)*2))  # each regressor can be TRUE or FALSE
    names(regMat) <- paste(rep(regressors, 2))

    # This assumes that there is a 'ScaledWaterYearIndex' column (true as it is automatically-generated)
    # Add loc and scale to names
    names(regMat)[1:length(regressors)] <- paste0(names(regMat)[1:length(regressors)], "_loc")
    names(regMat)[(1+length(regressors)):(length(regressors)*2)] <- paste0(names(regMat)[(1+length(regressors)):(length(regressors)*2)], "_scale")

    # Find which columns are not water year but are location / scale
    locIndex <- intersect(grep("WaterYear", names(regMat), invert = TRUE), grep("loc", names(regMat)))
    scaleIndex <- intersect(grep("WaterYear", names(regMat), invert = TRUE), grep("scale", names(regMat)))

    # Find which rows have more than one covariate as TRUE (excluding water year)
    # Keep only rows where fewer than two covariates other than water year are TRUE in both location and scale
    regMat2 <- regMat[(apply(regMat, 1, function(x) length(which(x[locIndex] == "TRUE"))) < 2 & apply(regMat, 1, function(x) length(which(x[scaleIndex] == "TRUE"))) < 2), ]

    # Find if both location and scale are TRUE in one covariate (excluding water year) and remove row if not the same covariate for both loc and scale
    for(i in 1:nrow(regMat2)){
      a = which(regMat2[i, locIndex] == "TRUE")  # which column (covariate) is TRUE for each row (corresponding to loc parameter)
      b = which(regMat2[i, scaleIndex] == "TRUE")  # which column (covariate) is TRUE for each row (corresponding to scale parameter)
      if(length(a) > 0 & length(b) > 0){  # if there is a TRUE for both loc and scale
        regMat2$Remove[i] <- ifelse(a != b, "Remove", "Keep")  # then check they are for the same covariate, if not, set to 'Remove'
      }
    }
    regMat3 <- regMat2[regMat2$Remove == "Keep", 1:(length(regressors)*2)]  # remove rows

    regMat <- regMat3
  }

  ############ OPTION 3: Create a matrix that only uses one covariate at a time in location and/or scale parameter ############
  if(option==3){
    # There will not be a different covariate in location and scale
    # This includes water year as a regressor
    regMatNames <- c(paste0(regressors, "_loc"), paste0(regressors, "_scale"))  # set names
    # Set up data table of FALSEs which has 2*number of regressors as number of columns and 3*number of regressors+1 as number of rows
    # so we can have TRUE in each regressor (loc and scale separately and loc and scale together) and then stationary model too
    regMat <- setNames(data.table(matrix(FALSE, nrow = 3*length(regressors)+1, ncol = length(regMatNames))), c(regMatNames))
    for(i in 1:length(regMatNames)){
      regMat[i,i] <- TRUE  # add TRUE for each regressor on new row for location and scale
      if(i < (3*length(regressors)+1)-length(regMatNames)){
        regMat[i+length(regMatNames), c(i, i+length(regressors))] <- TRUE  # make both loc and scale TRUE for each regressor
      }
    }
  }

  #############################################################################################################################

  # Number of regressors
  n <- length(regressors)

  # Formulas for all candidate models (first row is the location parameter, second row is the scale parameter)
  allModelsList <- apply(regMat, 1, function(x)
    c(paste(c("~ 1",regressors[x[1:n]]),
            collapse=" + "),
      paste(c("~ 1", regressors[x[(n+1):(2*n)]]),
            collapse=" + ")))

  # Fit all candidate models
  allModelsResults <- list()

  distFamily = as.name(tolower(dist))

  set.seed(1000)

  for(i in 1:ncol(allModelsList)){
    modelsList <<- allModelsList[, i]
    allModelsResults[[i]] <- suppressWarnings(fitWarning(Flow, all_data, distFamily, modelsList))
    if(!is.null(allModelsResults[[i]]$res)){
      allModelsResults[[i]]$res$AIC <- 2*length(allModelsResults[[i]]$res$par) + 2*(-allModelsResults[[i]]$res$loglik)
      allModelsResults[[i]]$res$BIC <- length(allModelsResults[[i]]$res$par)*log(length(allModelsResults[[i]]$res$data$y)) + 2*-allModelsResults[[i]]$res$loglik
    }
  }

  # Select 'best' model based on AIC or BIC (where model$res is not NULL)
  if(best_fit_method == "AIC"){
    best_indx <- which.min(lapply(allModelsResults[!sapply(allModelsResults, function(x) is.null(x$res))], function(x) x$res$AIC))
  } else if(best_fit_method == "BIC"){
    best_indx <- which.min(lapply(allModelsResults[!sapply(allModelsResults, function(x) is.null(x$res))], function(x) x$res$BIC))
  }

  # Model formula - first entry is the location parameter and second is the scale parameter
  best_model_formula <- allModelsList[, best_indx]
  modelsList <<- allModelsList[, best_indx]
  best_model <- suppressWarnings(fitWarning(Flow, all_data, distFamily, modelsList, methodType="bootstrap"))

  # Find models with warnings or errors
  warn_model_indx <- lapply(allModelsResults, function(x) length(x$warn) > 0)  # TRUE if warnings
  err_model_indx <- lapply(allModelsResults, function(x) length(x$err) > 0)  # TRUE if errors

  if(warn_model_indx[[best_indx]] == TRUE){
    print(paste0("There were warnings when fitting the best model based on ", best_fit_method, "."))
  }
  if(err_model_indx[[best_indx]] == TRUE){
    print(paste0("There were errors when fitting the best model based on ", best_fit_method, "."))
  }

  # Create a table containing the formulas used for fitting, any warnings or errors, AIC, BIC and rank
  fitting_metadata <- setNames(data.table(matrix(NA_character_, nrow = ncol(allModelsList), ncol = 4)), c("locFit", "scaleFit", "warnings", "errors"))
  fitting_metadata$BIC <- fitting_metadata$AIC <- NA_real_

  for(i in 1:ncol(allModelsList)){
    fitting_metadata$locFit[i] <- allModelsList[1,i]
    fitting_metadata$scaleFit[i] <- allModelsList[2,i]
    fitting_metadata$warnings[i] <- paste0(allModelsResults[[i]]$warn, collapse = " & ")
    fitting_metadata$errors[i] <- paste0(allModelsResults[[i]]$err, collapse = " & ")
    if(!is.null(allModelsResults[[i]]$res)){
      fitting_metadata$params[i] <- paste0(names(allModelsResults[[i]]$res$coefficients), " = ", round(allModelsResults[[i]]$res$coefficients, 5), collapse = "; ")
      fitting_metadata$AIC[i] <- allModelsResults[[i]]$res$AIC
      fitting_metadata$BIC[i] <- allModelsResults[[i]]$res$BIC
    } else{
      fitting_metadata$params[i] <- NA
      fitting_metadata$AIC[i] <- NA
      fitting_metadata$BIC[i] <- NA
    }
  }

  # Rank based on AIC
  fitting_metadata$RankAIC <- rank(fitting_metadata$AIC, na.last = "keep")
  # Rank based on BIC
  fitting_metadata$RankBIC <- rank(fitting_metadata$BIC, na.last = "keep")
  # Covariate used to detrend the data
  fitting_metadata$DetrendingCovariate <- ifelse(!is.null(detrend.cov), detrend.cov, NA)
  # Option
  fitting_metadata$CovariatesOption <- option

  # Select stationary model (last one in allModelsResults because of the way regMat is set up for all options)
  set.seed(1000)
  modelsList <<- allModelsList[,length(allModelsResults)]
  stat_model <- suppressWarnings(fitWarning(Flow, all_data, distFamily, modelsList, methodType="bootstrap"))

  # Check that there are only 3 parameters
  if(length(stat_model$res$par) > 3){
    stop("Stationary model has more than three parameters. Check this is actually the stationary model.")
  }

  return(list(best_model = best_model, best_model_formula = best_model_formula, stat_model = stat_model, all_models_metadata = fitting_metadata))
}


#' UserModel
#'
#' Internal function. Fits model selected by the user.
#'
#' @param dist distribution to use for fitting (GEV or GLO).
#' @param all_data_raw data.frame of response variable (1st column) and regressors (contains raw water year and WaterYearIndex for output files and plots).
#' @param locFit the regression equation including covariates to include in the location parameter (copy from csv containing all the model fits).
#' @param scaleFit the regression equation including covariates to include in the scale parameter (copy from csv containing all the model fits).
#' @param detrend.cov optional character string. If included, instructs to detrend based on covariate in that string (usually time).
#'
#' @return The following are returned:
#' @return user_model - details of user-selected model.
#' @return user_model_formula - details of the formula used to fit the model (should be the same as the locFit and scaleFit input).
#' @keywords internal
UserModel <- function(dist, all_data_raw, locFit, scaleFit, detrend.cov){

  all_data <- all_data_raw[, !names(all_data_raw) %in% c("WaterYear", "WaterYearIndex")]

  # Extract predictor variables - keep all covariates and exclude 'Flow' column
  covariateData <- all_data[, !names(all_data) %in% "Flow"]
  regressors <- names(covariateData)

  # Detrend covariates based on what has been specified in detrend.cov - should be time-based covariate
  # If detrend.cov is not null then detrend data
  if(!is.null(detrend.cov)) {
    all_data <- detrend.fn(all_data, detrend.cov, regressors)
  }

  # Reset covariateData now that the data have been detrended
  covariateData <- all_data[, !names(all_data) %in% "Flow"]

  # Fit model
  distFamily = as.name(tolower(dist))

  set.seed(1000)
  modelsList <<- c(locFit, scaleFit)
  modelResults <- suppressWarnings(fitWarning(Flow, all_data, distFamily, modelsList, method="bootstrap"))

  if(length(modelResults$warn) > 0){
    print("There were warnings when fitting the user-selected model.")
  }
  if(length(modelResults$err) > 0){
    print("There were errors when fitting the user-selected model.")
  }

  return(list(user_model = modelResults, user_model_formula = c(locFit, scaleFit)))
}


#' CalcCondRLs
#'
#' Internal function. Calculates conditional return levels for a given model.
#'
#' @param all_data_raw flow and covariate data (raw and scaled water year and not detrended covariates).
#' @param model fitted model (output of BestModel function or UserModel function).
#' @param ret_periods vector of return periods of interest.
#' @param CI confidence interval for results.
#'
#' @return The output is ret_levels_cond - a data.frame of return levels for each return period (includes covariates and flow data too)
#' @keywords internal
CalcCondRLs <- function(all_data_raw, model, ret_periods, CI) {

  # Set alpha based on confidence interval required (input to predict function)
  alpha <- 1 - CI/100

  # Calculate conditional return levels given observed covariates

  # Calculate return levels
  ret_levels <- predict(model, M=ret_periods, ci.fit=TRUE, alpha=alpha, unique.=FALSE)

  # Write out results

  # If stationary model (i.e. number of parameters is 3)
  if(length(model$map$par) == 3){

    # Set up flow column for results table
    resultsTable <- data.frame(WaterYear = all_data_raw$WaterYear, Flow = all_data_raw$Flow)

    # Add columns of results for every user-defined RP
    dataFrameCols <- paste0("\"", RPs, "-year (lower ", CI, "% CI)\" =
                            signif(unname(ret_levels$obj$M.", ret_periods, "[,3]),4),
                            \"", RPs, "-year estimate\" = signif(unname(ret_levels$obj$M.", ret_periods, "[,2]),4),
                            \"", RPs, "-year (upper ", CI, "% CI)\" = signif(unname(ret_levels$obj$M.", ret_periods, "[,4]),4)",
                            collapse = ",")

    # Add the above into a data frame
    eval(parse(text=paste0("ret_levels_cond <- data.frame(resultsTable, ", dataFrameCols, ")")))

    # For each user-defined RP create three column names (lower CI, estimate, upper CI)
    names <- paste0("\"", ret_periods, "-year (lower ", CI, "% CI)\", \"", ret_periods, "-year\", \"", ret_periods, "-year (upper ", CI, "% CI)\"", collapse = ",")

    # Set all column names incorporating the above
    eval(parse(text=paste0("names(ret_levels_cond) <- c(\"WaterYear\", \"Flow\", ", names, ")")))

  } else{  # if non-stationary

    # Select covariate data used in the model
    cov_data <- as.data.frame(do.call(cbind, model$map$data$D))
    covs <- cov_data[, unique(names(cov_data))] %>% select(-"(Intercept)")

    # Set up flow and covariates columns for results table
    # Include data (detrended if applicable) and water year both raw and scaled (and potentially detrended if detrend.cov = another covariate)
    resultsTable <- data.frame(covs)  # use the data in here as this will be the detrended/scaled data used in the analysis
    # Add flow, water year and water year index
    resultsTable$Flow <- all_data_raw$Flow
    resultsTable$WaterYear <- all_data_raw$WaterYear
    if("ScaledWaterYearIndex" %in% names(resultsTable)){
      resultsTable$WaterYearIndex <- all_data_raw$WaterYearIndex
      resultsTable <- resultsTable %>% select(WaterYear, WaterYearIndex, Flow, ScaledWaterYearIndex, everything())
    } else{
      resultsTable <- resultsTable %>% select(WaterYear, Flow, everything())
    }

    # Add columns of results for every user-defined RP
    dataFrameCols <- paste0("\"", ret_periods, "-year (lower ", CI, "% CI)\" = signif(unname(ret_levels$obj$M.", ret_periods, "[,3]),4),
                        \"", ret_periods, "-year estimate\" = signif(unname(ret_levels$obj$M.", ret_periods, "[,2]),4),
                        \"", ret_periods, "-year (upper ", CI, "% CI)\" = signif(unname(ret_levels$obj$M.", ret_periods, "[,4]),4)",
                            collapse = ",")

    # Add the above into a data frame
    eval(parse(text=paste0("ret_levels_cond <- data.frame(resultsTable, ", dataFrameCols, ")")))

    # For each user-defined RP create three column names (upper CI, estimate, upper CI)
    names <- paste0("\"", ret_periods, "-year (lower ", CI, "% CI)\", \"", ret_periods, "-year\", \"", ret_periods, "-year (upper ", CI, "% CI)\"", collapse = ",")

    # Set all column names incorporating the above
    # Need to do first columns separately to use covariate data
    eval(parse(text=paste0("names(ret_levels_cond)[(ncol(resultsTable)+1):ncol(ret_levels_cond)] <- c(", names, ")")))
  }

  return(ret_levels_cond)

}


#' marg_level_RP
#'
#' Internal function used in CalcMargRLs and BootstrapMargRLs (via bootMargRLfn). Calculates marginal
#' return levels by integrating over the covariate distribution.
#'
#' @param R return periods
#' @param params fitted model parameters (must include mu, phi and xi columns)
#' @param dimcov sample size of covariates
#' @param dist distribution type ("GEV" or "GLO")
#'
#' @return Returns data frame of marginal return levels for each return period.
#' @keywords internal
marg_level_RP <- function(R, params, dimcov, dist) {

  # Equation that must be solved (see Eastoe and Tawn 2009)
  marg_level <- function(q, params, dimcov, dist, R) {
    params = data.frame(params)
    mu = params$mu
    sigma = exp(params$phi)
    xi = params$xi
    if(dist == "GEV"){
      1-1/R - (1/dimcov)*sum(1-pgev(rep(q,dimcov), mu=mu, sigma=sigma, xi=xi, lower.tail=F))
    } else if(dist == "GLO"){
      1-1/R - (1/dimcov)*sum(1-pglo(rep(q,dimcov), mu=mu, sigma=sigma, xi=xi, lower.tail=F))
    }
  }

  # Solve for q (catches errors)
  warn <- err <- NA
  res <- withCallingHandlers(
    tryCatch(uniroot(f = marg_level, interval = c(0, 1e+06), R = R, params = params, dimcov = dimcov, dist = dist),
             error=function(e){
               err <<- conditionMessage(e)
               NULL  # NULL
             }), warning=function(w){
               warn <<- append(warn, conditionMessage(w))
             })
  result <- ifelse(!is.null(res), res$root, NA)
  list(res = result, warn = warn, err = err)
}


#' bootMargRLfn
#'
#' Internal function used in BootstrapMargRLs function. Estimates marginal return level
#' or present-day marginal return level depending on whether present_day is TRUE or FALSE.
#'
#' @param boot_model bootstrap model
#' @param covar_resamp resampled covariates (only used for present-day marginals)
#' @param ret_periods required return periods
#' @param dist distribution type ("GEV" or "GLO")
#' @param pres_day TRUE or FALSE
#'
#' @return Output is marginal return levels or present-day marginal return levels and associated warnings and errors.
#' @keywords internal
bootMargRLfn <- function(boot_model, covar_resamp, ret_periods, dist, pres_day){
  if(pres_day == FALSE){
    # 5. Get fitted parameters for boot_model
    boot_model_pars <- linearPredictors(boot_model$res, unique.=FALSE)$link
  } else if(pres_day == TRUE){
    # 5. Get fitted parameters for boot_model with time fixed in resampled covariates
    covar_resamp_pres <- covar_resamp
    covar_resamp_pres$ScaledWaterYearIndex <- max(covar_resamp$ScaledWaterYearIndex)

    boot_model_pars <- linearPredictors(boot_model$res, newdata=covar_resamp_pres, unique.=FALSE)$link
  }

  # 6. Estimate marginal RL
  dimcov = nrow(boot_model$res$data$D$mu)  # sample size of covariates
  boot_marg_RLs_res_warn_err <- sapply(ret_periods, marg_level_RP, params=boot_model_pars, dimcov=dimcov, dist=dist)
  boot_marg_RLs_res <- boot_marg_RLs_res_warn_err[which(row.names(boot_marg_RLs_res_warn_err) == "res"),]
  boot_marg_RLs_j <- do.call(cbind, boot_marg_RLs_res)

  # Report the warnings and errors - number reported is out of the number of return periods,
  # so if every RP has failed, it will say the number of RPs.
  boot_marg_RLs_warn <- boot_marg_RLs_res_warn_err[which(row.names(boot_marg_RLs_res_warn_err) == "warn"),]
  boot_marg_RLs_warnings_j <- do.call(cbind, boot_marg_RLs_warn)

  boot_marg_RLs_err <- boot_marg_RLs_res_warn_err[which(row.names(boot_marg_RLs_res_warn_err) == "err"),]
  boot_marg_RLs_errors_j <- do.call(cbind, boot_marg_RLs_err)

  return(list(boot_marg_RLs_j = boot_marg_RLs_j, boot_marg_RLs_warnings_j = boot_marg_RLs_warnings_j,
              boot_marg_RLs_errors_j = boot_marg_RLs_errors_j))
}


#' bootCIs
#'
#' Internal functon used in the BootstrapMargRLs function. Writes out warning messages associated with
#' the bootstrap models and calculates and returns the confidence intervals based on the bootstrapping.
#' Used for both marginal and present-day marginal return levels.
#'
#' @param ret_periods numeric vector of required return periods
#' @param boot_marg_ret_levels matrix containing the marginal (or present-day marginal) return levels for each bootstrap sample and return period
#' @param boot_marg_ret_levels_warnings matrix containing warnings associated with the above
#' @param boot_marg_ret_levels_errors matrix containing errors associated with the above
#' @param pres_day TRUE or FALSE
#'
#' @return The confidence intervals are returned for the marginal or present-day marginal return levels.
#' @keywords internal
bootCIs <- function(ret_periods, boot_marg_ret_levels, boot_marg_ret_levels_warnings, boot_marg_ret_levels_errors, pres_day){

  # Add comment to warning messages if reference is to present-day marginal bootstrapping
  pd_comment <- ifelse(pres_day == TRUE, " (present-day marginal) ", " ")

  # Check number of warnings for each bootstrap model
  num_warnings_RPs <- colSums(!is.na(boot_marg_ret_levels_warnings))
  if(any(num_warnings_RPs > 10)){
    print(paste0("There are warnings in more than 10 of the", pd_comment, "bootstrap models for at least one return period"))
  }

  # Check number of errors for each bootstrap model
  num_errors_RPs <- colSums(!is.na(boot_marg_ret_levels_errors))
  if(any(num_errors_RPs > 10)){
    print(paste0("There are errors in more than 10 of the", pd_comment, "bootstrap models for at least one return period"))
  }

  # Add column names to boot_marg_RPs for writing out - add after above check as otherwise the names don't match
  colnames(boot_marg_ret_levels) <- ret_periods

  # Calculate confidence interval - this should remove any NAs so important to check that there are not many (prints warning if >5% are NAs)
  ci = CI/100  # convert to decimal from percentage
  boot_quantiles <- sort(c((1-ci)/2, (1+ci)/2))
  boot_conf <- data.frame(t(apply(boot_marg_ret_levels, 2, quantile, boot_quantiles, na.rm = T)))
  names(boot_conf) <- boot_quantiles

  boot_out <- cbind(ret_periods, boot_conf)
  names(boot_out)[1] <- "RP"

  return(boot_out)

}


#' BootstrapMargRLs
#'
#' Internal function used in CalcMargRLs. Calculates confidence intervals for
#' marginal (and present-day marginal) return levels.
#'
#' @param dist distribution type ("GEV" or "GLO")
#' @param model fitted model
#' @param ret_periods numeric vector of required return periods
#' @param cov.data covariate data (detrended if applicable)
#' @param CI confidence interval for results as a percentage
#' @param present_day logical stating whether to calculate present-day marginal return level CI - TRUE or FALSE
#' @param bootstrap_n number of bootstrap samples for estimating confidence intervals
#'
#' @return The output includes the following:
#' @return boot_model_warnings contains the number of warnings per bootstrap model - written out to check maximum number and how many models are affected.
#' @return boot_model_errors contains the number of errors that occurred per bootstrap model - want this to be zero.
#' @return boot_out is the final confidence interval for the marginal return levels.
#' @return boot_marg_RLs contains the marginal return level results from all the bootstrap models (columns are return periods, rows are bootstrap samples).
#' @return boot_marg_RLs_warnings contains the warnings for each bootstrap model for each return period.
#' @return boot_marg_RLs_errors is the same but for errors.
#' @return Those names containing 'pres' are the same as above but for the present-day marginal return levels. Present-day outputs will be empty if present_day = FALSE.
BootstrapMargRLs <- function(dist, model, ret_periods, cov.data, CI, present_day, bootstrap_n) {

  # Initialise sample matrices and loop over number of bootstrap samples
  boot_marg_RLs <- boot_marg_RLs_warnings <- boot_marg_RLs_errors <- matrix(NA, nrow = bootstrap_n, ncol = length(ret_periods))

  boot_model_warnings <- boot_model_errors <- c()

  # Initialise matrices for present-day marginal return levels if present_day = TRUE
  if(present_day==TRUE){
    boot_pres_marg_RLs <- boot_pres_marg_RLs_warnings <- boot_pres_marg_RLs_errors <- matrix(NA, nrow = bootstrap_n, ncol = length(ret_periods))
  } else{
    boot_pres_marg_RLs <- boot_pres_marg_RLs_warnings <- boot_pres_marg_RLs_errors <- NULL
  }

  for (j in 1:bootstrap_n) {

    # 1. Resample covariates keeping time fixed
    time_col <- which(names(cov.data) == "ScaledWaterYearIndex")

    set.seed(j)
    indx <- sample(1:nrow(cov.data), replace = T)
    covar_resamp <- cov.data[indx,]

    # Keep time ordering in covar_resamp as time has no uncertainty
    if(length(time_col) > 0) {
      covar_resamp[, time_col] <- cov.data[, time_col]
    }

    # 2. Get new fitted parameters with the new resampled covariates
    boot_pars <- linearPredictors(model, newdata=covar_resamp, unique.=FALSE)$link

    # 3. Generate new time series using new parameters
    for (i in 1:length(covar_resamp$Flow)) {
      if(dist == "GEV"){
        covar_resamp$Flow[i] <- rgev(1, boot_pars[i, grep("mu", colnames(boot_pars))], exp(boot_pars[i, grep("phi", colnames(boot_pars))]),
                                     boot_pars[i, grep("xi", colnames(boot_pars))])
      } else if(dist == "GLO"){
        covar_resamp$Flow[i] <- rglo(1, boot_pars[i, grep("mu", colnames(boot_pars))], exp(boot_pars[i, grep("phi", colnames(boot_pars))]),
                                     boot_pars[i, grep("xi", colnames(boot_pars))])
      }
    }

    # 4. Refit model using resampled covariates
    distFamily = as.name(tolower(dist))
    modelsList <<- c(model$map$formulae$mu, model$map$formulae$phi)
    set.seed(1000)
    boot_model <- suppressWarnings(fitWarning(Flow, covar_resamp, distFamily, modelsList))

    # Print message if warnings and errors
    if(length(boot_model$warn) > 1){
      print(paste0("There were ", length(boot_model$warn), " warnings when fitting the model for bootstrap sample ", j, "."))
    } else if(length(boot_model$warn) == 1){
      print(paste0("There was ", length(boot_model$warn), " warning when fitting the model for bootstrap sample ", j, "."))
    }

    if(length(boot_model$err) > 1){
      print(paste0("There were ", length(boot_model$err), " errors when fitting the model for bootstrap sample ", j, ". This bootstrap sample will not be used to calculate confidence intervals."))
    } else if(length(boot_model$err) == 1){
      print(paste0("There was ", length(boot_model$err), " error when fitting the model for bootstrap sample ", j, ". This bootstrap sample will not be used to calculate confidence intervals."))
    }

    boot_model_warnings[j] <- length(boot_model$warn)
    boot_model_errors[j] <- length(boot_model$err)

    # If boot_model has worked and not produced an error then complete the calculations
    if(!is.null(boot_model$res)){
      # 5 and 6. See bootMargRLfn.
      # Marginal return level
      bootMargRL <- bootMargRLfn(boot_model, covar_resamp = NULL, ret_periods, dist, pres_day = FALSE)

      boot_marg_RLs[j,] <- bootMargRL$boot_marg_RLs_j
      boot_marg_RLs_warnings[j,] <- bootMargRL$boot_marg_RLs_warnings_j
      boot_marg_RLs_errors[j,] <- bootMargRL$boot_marg_RLs_errors_j

    } else{  # else write out NAs
      boot_marg_RLs[j,] <- boot_marg_RLs_warnings[j,] <- boot_marg_RLs_errors[j,] <- NA
    }

      # Present-day marginal return level
    if(present_day == TRUE){
      if(!is.null(boot_model$res)){
        bootPresMargRL <- bootMargRLfn(boot_model, covar_resamp, ret_periods, dist, pres_day = TRUE)

        boot_pres_marg_RLs[j,] <- bootPresMargRL$boot_marg_RLs_j
        boot_pres_marg_RLs_warnings[j,] <- bootPresMargRL$boot_marg_RLs_warnings_j
        boot_pres_marg_RLs_errors[j,] <- bootPresMargRL$boot_marg_RLs_errors_j
      } else{
        boot_pres_marg_RLs[j,] <- boot_pres_marg_RLs_warnings[j,] <- boot_pres_marg_RLs_errors[j,] <- NA
      }
    }

  }

  # Write message if there are warnings or errors in more than 10 of the bootstrap models for at least one return period
  # and calculate confidence intervals

  # Marginal return level
  boot_out <- bootCIs(ret_periods, boot_marg_RLs, boot_marg_RLs_warnings, boot_marg_RLs_errors, pres_day = FALSE)

  # Present-day marginal return level
  if(present_day == TRUE){
    boot_pres_out <- bootCIs(ret_periods, boot_pres_marg_RLs, boot_pres_marg_RLs_warnings, boot_pres_marg_RLs_errors, pres_day = TRUE)
  } else{
    boot_pres_out <- NULL
  }

  return(list(boot_model_warnings = boot_model_warnings, boot_model_errors = boot_model_errors, boot_out = boot_out,
              boot_marg_RLs = boot_marg_RLs, boot_marg_RLs_warnings = boot_marg_RLs_warnings, boot_marg_RLs_errors = boot_marg_RLs_errors,
              boot_pres_out = boot_pres_out, boot_pres_marg_RLs = boot_pres_marg_RLs, boot_pres_marg_RLs_warnings = boot_pres_marg_RLs_warnings,
              boot_pres_marg_RLs_errors = boot_pres_marg_RLs_errors))
}


#' CalcMargRLs
#'
#' Internal function. Calculates marginal (and present-day marginal) return levels
#' by integrating over the covariate distribution.
#'
#' @param dist distribution type ("GEV" or "GLO")
#' @param model fitted model (output of BestModel function or UserModel function)
#' @param all_data_raw flow and covariate data (raw and scaled water year and not detrended covariates)
#' @param ret_periods numeric vector of required return periods
#' @param CI confidence interval for results, e.g. 90 for 90\% confidence interval
#' @param detrend.cov as in BestModel function
#' @param present_day logical stating whether to calculate present-day marginal return level CI - TRUE or FALSE
#' @param bootstrap_n number of bootstrap samples for estimating confidence intervals
#'
#' @return Marginal and present-day marginal return levels for each return period are included in the output list, along with warnings and errors.
#' @keywords internal
CalcMargRLs <- function(dist, model, all_data_raw, ret_periods, CI, detrend.cov, present_day, bootstrap_n = 200) {

  # Covariate data (covariates used and saved in the model are detrended (if user has selected
  # this option) so need to detrend covariates here too)
  all_data <- all_data_raw[, !names(all_data_raw) %in% c("WaterYear", "WaterYearIndex")]

  # Extract predictor variables - keep all covariates and exclude 'Flow' column
  covariateData <- all_data[, !names(all_data) %in% "Flow"]
  regressors <- names(covariateData)

  # Detrend covariates based on what has been specified in detrend.cov - should be time-based covariate
  # If detrend.cov is not null
  if(!is.null(detrend.cov)) {
    all_data <- detrend.fn(all_data, detrend.cov, regressors)
  }

  cov.data <- all_data

  # Marginal return levels
  # Find parameters of model
  model_pars <- linearPredictors(model, unique.=FALSE)$link
  # Sample size of covariates
  dimcov = nrow(model$map$data$D$mu)

  # Estimate of marginal return level
  RL_warn_err <- sapply(ret_periods, marg_level_RP, params = model_pars, dimcov = dimcov, dist = dist)
  RL <- RL_warn_err[which(row.names(RL_warn_err) == "res"),]
  RLdf <- t(do.call(cbind.data.frame, RL))
  ret_levels_marg <- data.frame(RP = ret_periods, RL = RLdf)
  rownames(ret_levels_marg) <- NULL

  # CIs (will calculate for present-day marginal return levels too if present_day = TRUE)
  bootstrap_results <- BootstrapMargRLs(dist, model, ret_periods, cov.data, CI, present_day, bootstrap_n)

  # Marginal return level results
  marg_ret_levels_CIs <- bootstrap_results$boot_out
  marg_ret_levels_out <- data.frame(RP = paste0(ret_levels_marg$RP, "-year"), lower_CI = marg_ret_levels_CIs[,2], estimates = ret_levels_marg$RL,
                                    upper_CI = marg_ret_levels_CIs[,3])
  names(marg_ret_levels_out) <- c("RP", paste0("Lower ", CI, "% CI"), "Estimate", paste0("Upper ", CI, "% CI"))

  # Estimate of present-day marginal return level (if present_day = TRUE)
  if(present_day == TRUE){
    # 1. Resample covariates and fix time to be the last year
    set.seed(1000)
    indx <- sample(1:nrow(cov.data), replace = T)
    cov_resamp <- cov.data[indx,]
    cov_resamp$ScaledWaterYearIndex <- max(cov.data$ScaledWaterYearIndex)

    # 2. Get new fitted parameters with the new resampled covariates
    pres_pars <- linearPredictors(model, newdata=cov_resamp, unique.=FALSE)$link

    # 3. Estimate present-day marginal RL
    pres_marg_RLs_res_warn_err <- sapply(ret_periods, marg_level_RP, params = pres_pars, dimcov = dimcov, dist = dist)
    pres_RL <- pres_marg_RLs_res_warn_err[which(row.names(pres_marg_RLs_res_warn_err) == "res"),]
    pres_RLdf <- t(do.call(cbind.data.frame, pres_RL))
    pres_ret_levels_marg <- data.frame(RP = ret_periods, RL = pres_RLdf)
    rownames(pres_ret_levels_marg) <- NULL

    # Present-day marginal return level results
    pres_marg_ret_levels_CIs <- bootstrap_results$boot_pres_out
    pres_marg_ret_levels_out <- data.frame(RP = paste0(pres_ret_levels_marg$RP, "-year"), lower_CI = pres_marg_ret_levels_CIs[,2],
                                           estimates = pres_ret_levels_marg$RL, upper_CI = pres_marg_ret_levels_CIs[,3])
    names(pres_marg_ret_levels_out) <- c("RP", paste0("Lower ", CI, "% CI"), "Estimate", paste0("Upper ", CI, "% CI"))
  } else{
    pres_marg_ret_levels_out = NULL

  }

  return(list(marg_ret_levels_out = marg_ret_levels_out, boot_model_warnings = bootstrap_results$boot_model_warnings,
              boot_model_errors = bootstrap_results$boot_model_errors, boot_marg_RLs = bootstrap_results$boot_marg_RLs,
              boot_marg_RLs_warnings = bootstrap_results$boot_marg_RLs_warnings,
              boot_marg_RLs_errors = bootstrap_results$boot_marg_RLs_errors, pres_marg_ret_levels_out = pres_marg_ret_levels_out,
              boot_pres_marg_RLs = bootstrap_results$boot_pres_marg_RLs, boot_pres_marg_RLs_warnings = bootstrap_results$boot_pres_marg_RLs_warnings,
              boot_pres_marg_RLs_errors = bootstrap_results$boot_pres_marg_RLs_errors, bootstrap_n = bootstrap_n))
}


#' CalcStatRLs
#'
#' Internal function. Calculates return levels of stationary model.
#'
#' @param dist distribution type ("GEV" or "GLO")
#' @param model stat_model output of BestModel function
#' @param ret_periods numeric vector of required return periods
#' @param CI confidence interval for results, e.g. 90 for 90\% confidence interval
#'
#' @return Output is ret_levels_stat - a data frame of return levels for each return period.
#' @keywords internal
CalcStatRLs <- function(dist, model, ret_periods, CI) {

  # Set alpha based on confidence interval required
  alpha <- 1 - CI/100

  # Calculate return levels (including CIs)
  RLs_stat <- predict(model, M = ret_periods, ci.fit = TRUE, alpha = alpha)
  RLs_stat_combined <- do.call(rbind, RLs_stat$obj)
  ret_levels_stat <- data.frame(RP = paste0(word(names(RLs_stat$obj), 2, sep = fixed(".")), "-year"),
                                LCI = RLs_stat_combined[,3], Estimate = RLs_stat_combined[,2], UCI = RLs_stat_combined[,4], row.names = NULL)
  names(ret_levels_stat) <- c("RP", paste0("Lower ", CI, "% CI"), "Estimate", paste0("Upper ", CI, "% CI"))

  return(ret_levels_stat)
}


#' CovariateModel
#'
#' Internal function. Calculates results for either the 'best fit' model or a
#' user-selected non-stationary model and also for the stationary model.
#'
#' @param dist distribution type ("GEV" or "GLO")
#' @param all_data_raw input data (as in BestModel function)
#' @param ret_periods numeric vector of required return periods
#' @param CI confidence interval as a percentage, e.g. input '90' for the 90\% confidence interval
#' @param model_fits output from running BestModel function
#' @param model_type either "Best" or "User". Select whether to use the best fit model based on AIC/BIC or the user-selected model to calculate return levels and plot.
#' @param userModel output from userModel function or NULL depending on above selection
#' @param locFit the regression equation including covariates to include in the location parameter (copy from metadata csv containing all the model fits from running BestModel function). Only required if model_type = "User".
#' @param scaleFit the regression equation including covariates to include in the scale parameter (copy from metadata csv containing all the model fits from running BestModel function). Only required if model_type = "User".
#' @param gaugeNumber number of the gauge to analyse
#' @param gaugeName name of the gauge to analyse
#' @param detrend.cov as in BestModel function. Only required if model_type = "User", otherwise outputs from BestModel are used.
#' @param present_day logical stating whether to calculate present-day marginal return level and its confidence interval - TRUE or FALSE
#'
#' @return A list containing a range of outputs is returned:
#' @return model - model used to calculate results - either the 'best fit' model based on AIC or BIC (as selected by the user in BestModel function) or the user-selected model (does not include warnings or errors).
#' @return model_formula - details of the formula used to fit the model.
#' @return condRLs - conditional return levels.
#' @return margRLs - marginal return levels.
#' @return pres_margRLs - present-day marginal return levels.
#' @return margRLsAllInfo - output from CalcMargRLs.
#' @return stat_model - details of stationary model.
#' @return statRLs - stationary return levels.
#' @return allModelsMetadata - same as the output from the BestModel function.
#' @keywords internal
CovariateModel <- function(dist, all_data_raw, ret_periods, CI, model_fits, model_type,
                           userModel, locFit, scaleFit, gaugeNumber, gaugeName, detrend.cov, present_day) {

  # Output of BestModel function
  modelFits = model_fits

  # Select whether to use best fitting model or user-selected model for calculating
  # return levels and creating plots and find the formulas used in the location and scale parameters
  if(model_type == "Best"){
    model <- modelFits$best_model$res
    model_formula <- modelFits$best_model_formula
  } else if(model_type == "User"){
    model <- userModel$user_model$res
    model_formula <- userModel$user_model_formula
  }

  # Check that the model contains ScaledWaterYearIndex if present_day = TRUE. If not, set present_day
  # to FALSE and print warning message.
  if(present_day == TRUE & length(grep("ScaledWaterYearIndex", names(model$map$coefficients)))==0){
    print("The option to calculate present_day is set to TRUE but the time (i.e. ScaledWaterYearIndex) covariate is not included in the selected model. Therefore, present_day has been set to FALSE.")

    present_day = FALSE
  }

  # If selected model is stationary then set present_day to FALSE
  if(length(model$map$par) == 3) present_day = FALSE

  # Diagnostic plots
  # Set up save directory
  savePath <- ifelse(nchar(gaugeName) > 0, paste0("Outputs/Physical covariates models/", gaugeNumber, " (", gaugeName, ")"),
                     paste0("Outputs/Physical covariates models/", gaugeNumber))
  outPath <- paste(savePath, dist, "model diagnostic plots.png")

  gaugeData <- data.frame(Flow = all_data_raw$Flow, WaterYear = all_data_raw$WaterYear,
                          WaterYearIndex = all_data_raw$WaterYearIndex, ScaledWaterYearIndex = all_data_raw$ScaledWaterYearIndex)
  model.diag.warn.err <- plot.diagnostic.warn.err(dist, model, outPath, gaugeNumber, gaugeName, gaugeData)

  # Calculate conditional return levels
  if(length(model$map$par) > 3){
    condRLs <- CalcCondRLs(all_data_raw, model, ret_periods, CI)
  } else{
    condRLs <- matrix("The selected model is stationary. Conditional return levels are not available for stationary models.")
  }


  # Calculate marginal return levels if model is not stationary
  if(length(model$map$par) > 3){
    margRLsAllInfo <- CalcMargRLs(dist, model, all_data_raw, ret_periods, CI, detrend.cov, present_day)
    margRLs <- margRLsAllInfo$marg_ret_levels_out
    if(present_day == TRUE){
      pres_margRLs <- margRLsAllInfo$pres_marg_ret_levels_out
    } else{
      pres_margRLs <- NULL
    }
  } else{
    margRLs <- matrix("The selected model is stationary. Marginal return levels are not available for stationary models.")
    pres_margRLs <- NULL
    margRLsAllInfo <- NULL
  }

  # Calculate stationary model return levels
  statRLs <- CalcStatRLs(dist, modelFits$stat_model$res, ret_periods, CI)

  # Set up a list of outputs from the function
  output <- list(model = model, model_formula = model_formula, condRLs = condRLs, margRLs = margRLs, pres_margRLs = pres_margRLs,
                 margRLsAllInfo = margRLsAllInfo, stat_model = modelFits$stat_model, statRLs = statRLs,
                 allModelsMetadata = modelFits$all_models_metadata)

  return(output)

}


#' PlotFit
#'
#' Internal function. Plots relationships between AMAX flow, covariates and results.
#'
#' @param dist distribution type ("GEV" or "GLO")
#' @param fit output of CovariateModel
#' @param ret_periods return periods to plot - only RPs in condRLs, margRLs and statRLs will be plotted
#' @param all_data_raw input data (as in BestModel function)
#' @param CI confidence interval as a percentage, e.g. input '90' for the 90\% confidence interval
#' @param gaugeNumber gauge number
#' @param gaugeName gauge name
#' @param detrend.cov as in BestModel function
#'
#' @return Three plots are returned. The first is flow plotted against all covariates.
#' @return The second is flow plotted against both conditional return levels and the stationary model return levels for the observed period.
#' @return The third is an encounter probability plot of the marginal and stationary results.
#' @keywords internal
PlotFit <- function(dist, fit = fit_results, ret_periods, all_data_raw, CI, gaugeNumber, gaugeName, detrend.cov) {

  # Set items for plot titles
  dist <- fit$model$map$family$name
  covs <- fit$model$map$formulae
  loc <- Reduce(paste, deparse(covs$mu))  # convert formula to string
  scale <- Reduce(paste, deparse(covs$phi))  # convert formula to string
  fitType <- ifelse(loc == "~1" & scale == "~1", "Stationary", "Non-stationary")
  covsInfo <- paste0("Location: ", as.character(loc), "; Scale: ", as.character(scale))


  # Plot 1 - plots relationships between flow and covariates
  all_data <- all_data_raw[, !names(all_data_raw) %in% c("WaterYear", "WaterYearIndex")]

  # Extract predictor variables - keep all covariates and exclude 'Flow' column
  covariateData <- all_data[, !names(all_data) %in% "Flow"]
  regressors <- names(covariateData)

  # Detrend covariates based on what has been specified in detrend.cov - should be time-based covariate
  # If detrend.cov is not null
  if(!is.null(detrend.cov)) {
    all_data <- detrend.fn(all_data, detrend.cov, regressors)
  }

  # Convert last column from a matrix to a numeric column
  all_data$ScaledWaterYearIndex <- as.numeric(all_data$ScaledWaterYearIndex)

  # Re-shape data
  covDataReshaped <- melt(all_data, id.var="Flow")

  # Set title
  plot1Title <- ifelse(nchar(gaugeName) > 0, paste0("Gauging station ", gaugeNumber, " (", gaugeName, ") \n Correlations between flow and covariate data"),
                 paste0("Gauging station ", gaugeNumber, " \n Correlations between flow and covariate data"))

  # Plot
  f <- ggplot(covDataReshaped, aes(x = Flow, y = value)) + geom_point(aes(color = variable)) +
    facet_grid(variable ~ ., scales = "free_y") + theme(legend.position = "none") + labs(x = expression(paste("Flow (", m^3, "/s)")), y = "Value") +
    ggtitle(plot1Title) + theme(plot.title = element_text(hjust = 0.5))


  # Plot 2 - plots conditional return levels for model used in CovariateModel
  #          as well as stationary return levels

  if(length(ret_periods) > 8) print("Only the first eight return periods will be plotted in the 'model results' plot.")
  lastRP <- min(length(ret_periods), 8)
  ret_periods_sub = ret_periods[1:lastRP]

  # Sort data for stationary model
  statRLsData <- data.frame(RP = fit$statRLs$RP, RL = fit$statRLs$Estimate)
  statRLsData <- statRLsData[which(word(statRLsData$RP, 1, sep = "-") %in% ret_periods_sub),]

  # If selected model is stationary
  if(fit$condRLs[1,1] == "The selected model is stationary. Conditional return levels are not available for stationary models."){

    statRLsData$RP2 <- as.numeric(word(statRLsData$RP, 1, sep = "-"))
    statRLsData$RP  <- with(statRLsData, reorder(RP, RP2))

    # Set title
    plot2Title <- ifelse(nchar(gaugeName) > 0, paste0("Gauging station ", gaugeNumber, " (", gaugeName, ") (", dist, ") \n Selected model: ",
                                                      covsInfo, " \n Stationary flow estimates"),
                         paste0("Gauging station ", gaugeNumber, " (", dist, ") \n Selected model: ",
                                covsInfo, " \n Stationary flow estimates"))

    # Plot
    g <- ggplot() + geom_segment(data = all_data_raw, aes(x = WaterYear, y = Flow, yend = 0, xend = WaterYear), size=2) +
      geom_hline(data = statRLsData, aes(yintercept = RL, color = RP), size = 1, linetype = "dashed") +
      scale_colour_brewer(name = "Return Period", palette = "Dark2") + labs(x = "Water year", y = expression(paste("Flow (", m^3, "/s)"))) +
      ggtitle(plot2Title) + theme(plot.title = element_text(hjust = 0.5))

  } else{  # if selected model is non-stationary
    # Conditional return levels
    CRLsData = fit$condRLs
    CRLsData <- CRLsData[, grep("CI", names(CRLsData), invert = TRUE)]  # to remove CIs
    CRLsData <- CRLsData[, grep("Year|year", names(CRLsData), value = TRUE)]
    CRLsData <- CRLsData[, grep("Index", names(CRLsData), value = TRUE, invert = TRUE)]
    condRLsData <- melt(CRLsData, id.var="WaterYear")
    names(condRLsData)[2] <- "RP"
    condRLsData <- condRLsData[which(word(condRLsData$RP, 1, sep = "-") %in% ret_periods_sub),]

    # Set title
    plot2Title <- ifelse(nchar(gaugeName) > 0, paste0("Gauging station ", gaugeNumber, " (", gaugeName, ") (", dist, ") \n Selected model: ",
                                                      covsInfo, " \n Stationary and conditional flow estimates"),
                         paste0("Gauging station ", gaugeNumber, " (", dist, ") \n Selected model: ",
                                covsInfo, " \n Stationary and conditional flow estimates"))

    g <- ggplot() + geom_segment(data = fit$condRLs, aes(x = WaterYear, y = Flow, yend = 0, xend = WaterYear), size=2) +
      geom_point(data = condRLsData, aes(x = WaterYear, y = value, color = RP)) +
      geom_hline(data = statRLsData, aes(yintercept = RL, color = RP), size = 1, linetype = "dashed") +
      scale_colour_brewer(name = "Return Period", palette = "Dark2") + labs(x = "Water year", y = expression(paste("Flow (", m^3, "/s)"))) +
      ggtitle(plot2Title) + theme(plot.title = element_text(hjust = 0.5))
  }

  # Plot 3 - Encounter probability plot
  FlowData <- data.frame(Dummy = rep(0, length(fit$model$map$data$y)), Flow = fit$model$map$data$y)
  N <- length(fit$model$map$data$y)  # length of flow data

  # Set title
  plot3Title <- ifelse(nchar(gaugeName) > 0, paste0("Gauging station ", gaugeNumber, " (", gaugeName, ") (", dist, ") \n Selected model: ",
                                                    covsInfo, " \n Encounter probability plot"),
                       paste0("Gauging station ", gaugeNumber, " (", dist, ") \n Selected model: ",
                              covsInfo, " \n Encounter probability plot"))

  # If the model is stationary
  if(fit$margRLs[1,1] == "The selected model is stationary. Marginal return levels are not available for stationary models."){
    statRLs = fit$statRLs
    names(statRLs)[2:4] <- c("StatLCI", "StatEstimate", "StatUCI")

    statRLs$RPnum <- as.numeric(word(statRLs$RP, 1, sep = fixed("-")))
    statRLs <- statRLs[order(statRLs$RPnum),]
    # Calculate encounter probability
    statRLs$EP <- 1-(1-1/as.numeric(statRLs$RPnum))^N

    statRLsData <- melt(statRLs[, which(! names(statRLs) %in% c("RP", "RPnum"))], id.var="EP")
    statRLsData$variable2 <- as.character(statRLsData$variable)
    statRLsData$variable2[statRLsData$variable2 %in% c("StatLCI", "StatUCI")] <- paste0(CI, "% confidence interval")
    statRLsData$variable2 <- as.factor(statRLsData$variable2)
    statRLsData$variable2 <- factor(statRLsData$variable2, levels(statRLsData$variable2)[c(3,2,1)])

    h <- ggplot() + geom_line(data = statRLsData, aes(x = EP, y = value, size=variable, linetype=variable2)) +
      labs(x = paste0("Encounter probability over the ", N, "-year record period"), y = expression(paste("Flow (", m^3, "/s)"))) +
      geom_rug(data = FlowData, aes(y=Flow)) +
      scale_size_manual(values = c(1,1.5,1)) + theme(legend.position = "none") +
      scale_linetype_manual(values = c("solid", "dashed"), labels = c("Stationary flow estimates", paste0(CI, "% confidence interval for the stationary estimates"))) +
      theme(legend.position = "bottom", legend.direction = "vertical", legend.title = element_blank()) + guides(size = FALSE) +
      ggtitle(plot3Title) + theme(plot.title = element_text(hjust = 0.5))

  } else{
    margRLs <- fit$margRLs
    names(margRLs)[2:4] <- c("MargLCI", "MargEstimate", "MargUCI")

    # Stationary RLs - need to recreate here to maintain all RPs, rather than subset used in Plot 2
    statRLs <- data.frame(RP = fit$statRLs$RP, RL = fit$statRLs$Estimate)
    names(statRLs)[2] <- "StatEstimate"

    statMarg <- merge(statRLs, margRLs, by = "RP")
    statMarg$RPnum <- as.numeric(word(statMarg$RP, 1, sep = fixed("-")))
    statMarg <- statMarg[order(statMarg$RPnum),]
    # Calculate encounter probability
    statMarg$EP <- 1-(1-1/as.numeric(statMarg$RPnum))^N

    statMargData <- melt(statMarg[, which(! names(statMarg) %in% c("RP", "RPnum"))], id.var="EP")
    statMargData$variable2 <- as.character(statMargData$variable)
    statMargData$variable2[statMargData$variable2 %in% c("MargLCI", "MargUCI")] <- paste0(CI, "% confidence interval")
    statMargData$variable2 <- as.factor(statMargData$variable2)
    statMargData$variable2 <- factor(statMargData$variable2, levels(statMargData$variable2)[c(3,2,1)])

    h <- ggplot() + geom_line(data = statMargData, aes(x = EP, y = value, size=variable, colour = variable2, linetype=variable2)) +
      labs(x = paste0("Encounter probability over the ", N, "-year record period"), y = expression(paste("Flow (", m^3, "/s)"))) +
      geom_rug(data = FlowData, aes(y=Flow)) +
      scale_colour_manual(values=c("#000000", "#0072B2", "#0072B2"), labels = c("Stationary flow estimates", "Integrated flow estimates (non-stationary)", paste0(CI, "% confidence interval for the non-stationary estimates"))) +
      scale_size_manual(values = c(1.5,1,1.5,1)) + theme(legend.position = "none") +
      scale_linetype_manual(values = c("solid","solid","dashed"), labels = c("Stationary flow estimates", "Integrated flow estimates (non-stationary)", paste0(CI, "% confidence interval for the non-stationary estimates"))) +
      theme(legend.position = "bottom", legend.direction = "vertical", legend.title = element_blank()) + guides(size = FALSE) +
      ggtitle(plot3Title) + theme(plot.title = element_text(hjust = 0.5))
  }

  output <- list(f = f, g = g, h = h)

  return(output)
}
