#' plot.selected.distributions
#'
#' This function plots the frequency analysis results from the fitted stationary distribution and one of the fitted non-stationary
#' distributions selected by the user (e.g. the best fitting non-stationary model based on likelihood ratio testing)
#' for one user-selected gauge using the outputs from fit.time for up to five return periods.
#' It also plots the confidence interval for one return period on a second plot. These plots will be automatically
#' written out to a .png file in the 'Outputs/Time covariate models' folder.
#'
#' @param folder string containing a path to a folder containing two sub-folders called 'Inputs' and 'Outputs'.
#' @param gauge string of the gauge number to plot. Gauge must have been analysed in the 'fit.time' function.
#' @param type the type of non-stationary distribution to plot. Choose from 'Varying location', 'Varying scale' and 'Varying location and scale'.
#' @param fitResults the output from the fitting function fit.time. This could either be from a saved location or may already be in the R environment if fit.time has just been run and the outputs saved to a named object.
#' @param RPs vector of return periods to plot. There must be a maximum of five selected.
#' @param CI.RP numeric (single). Return period for which a confidence interval is to be plotted. Must be in RPs.
#' @param CI numeric (single). The required confidence interval as a percentage to plot for the frequency analysis results from the distribution fitting, e.g. ‘90’ for 90\% is the default.
#' @param plotTitle optional parameter. String containing title for output plot if you wish to change the automatically-generated title. To split a title over two lines use ‘\\n’ where you want the line break to be.
#'
#' @return A .png file containing two plots is returned. The first shows the results from the fitted stationary and one user-selected non-stationary distribution for specified return periods. The second shows confidence intervals for one specified return period.
#' @export
#'
#' @examples
#' # Set the inputs to the function
#' folder <- "C:/Non-stationarity/Data"
#' gauge <- "24004"
#' type <- "Varying location"  # select a non-stationary distribution
#' fit.results <- readRDS("Outputs/Time covariate models/GEV fits from fit time function.RDS") # if saved the output from running the fit.time function OR
#' fit.results <- GEVfits  # if the output from the fit.time function is still in the R environment
#' RPs <- c(20, 50, 100)  # return periods to plot
#' CI.RP <- 100  # return period for confidence interval plot
#' CI <- 90  # set the confidence interval to be 90%
#' plot.selected.distributions(folder, gauge, type, fit.results, RPs, CI.RP, CI)
plot.selected.distributions <- function(folder, gauge, type, fitResults, RPs, CI.RP, CI=90, plotTitle = NULL){
  gaugeNumber <- gauge
  distributionType <- type

  alpha <- 1 - CI/100

  # If the user has set a plot title then use this
  if (!is.null(plotTitle)) title <- plotTitle

  # Set the working directory
  setwd(folder)

  # Check for existence of Input and Output folders and stop and print warning if missing
  if(!file.exists("Inputs") | !file.exists("Outputs")) stop("Inputs and/or Outputs folder does not exist", call. = FALSE)
  # Check that distribution type is valid
  if(! (distributionType == "Varying location" || distributionType == "Varying scale" || distributionType == "Varying location and scale")) stop("distributionType is not a valid option", call. = FALSE)
  # Check that there are only five RPs
  if(length(RPs) > 5) stop("A maximum of five return periods should be selected", call. = FALSE)
  # Check that CI.RP is in RPs
  if(!CI.RP %in% RPs) stop("Check that CI.RP is in RPs", call. = FALSE)
  # Check that gauge number is in fitResults
  if(!gaugeNumber %in% names(fitResults)) stop("Gauge number does not exist in fitResults", call. = FALSE)

  # Extract pre-processed gauge data from fit.time
  gaugeData <- eval(parse(text=paste0("fitResults$`", gaugeNumber, "`$data")))
  # Extract gauge name
  gaugeName <- gaugeData$GaugeName[1]
  # Extract distribution used in the fitting
  dist <- eval(parse(text=paste0("fitResults$`", gaugeNumber, "`$dist")))

  # Extract the stationary distribution fit from the fitting results
  statFit <- eval(parse(text=paste0("fitResults$'", gaugeNumber, "'$fit.stat")))
  # Extract type of non-stationary distribution selected by user to plot
  type <- ifelse(distributionType == "Varying location", "var.loc", ifelse(distributionType == "Varying scale", "var.scale", "var.loc.scale"))
  nonstatFit <- eval(parse(text=paste0("fitResults$'", gaugeNumber, "'$fit.", type)))

  # Set up the results table
  resultsTable <- data.frame(WaterYear = gaugeData$WaterYear, AMAX = gaugeData$Flow, WaterYearIndex = gaugeData$WaterYearIndex, ScaledWaterYearIndex = gaugeData$ScaledWaterYearIndex)

  # Find estimates and confidence intervals

  # Stationary
  stat.return.levels <- predict(statFit, M = RPs, ci.fit = TRUE, alpha = alpha, unique.=FALSE)

  # Non-stationary
  nonstat.return.levels <- predict(nonstatFit, M = RPs, ci.fit = TRUE, alpha = alpha, unique.=FALSE)

  # Add return level results to results table
  # Add stationary estimates first - these results are in a slightly different format from the non-stationary ones
  # Add columns of results for every user-defined RP
  dataFrameColsStat <- paste0("\"", RPs, "-year (lower ", CI, "% CI)\" =
                        signif(unname(stat.return.levels$obj$M.", RPs, "[,3]),4),
                        \"", RPs, "-year estimate\" = signif(unname(stat.return.levels$obj$M.", RPs, "[,2]),4),
                        \"", RPs, "-year (upper ", CI, "% CI)\" = signif(unname(stat.return.levels$obj$M.", RPs, "[,4]),4)",
                          collapse = ",")

  # Add the above into a data frame with the gauge data and the type of fit
  eval(parse(text=paste0("RPestimates.stat <- data.frame(resultsTable, ", dataFrameColsStat, ")")))

  # Add non-stationary estimates
  # Add columns of results for every user-defined RP
  dataFrameColsNonstat <- paste0("\"", RPs, "-year (lower ", CI, "% CI)\" = signif(unname(nonstat.return.levels$obj$M.", RPs, "[,3]),4),
                            \"", RPs, "-year estimate\" = signif(unname(nonstat.return.levels$obj$M.", RPs, "[,2]),4),
                            \"", RPs, "-year (upper ", CI, "% CI)\" = signif(unname(nonstat.return.levels$obj$M.", RPs, "[,4]),4)",
                                 collapse = ",")
  # Add the above into a data frame with the type of fit
  eval(parse(text=paste0("RPestimates.nonstat <- data.frame(resultsTable, ", dataFrameColsNonstat, ")")))

  # For each user-defined RP create three column names
  namesRP <- paste0("\"", RPs, "-year (lower ", CI, "% CI)\", \"", RPs, "-year estimate\", \"", RPs, "-year (upper ", CI, "% CI)\"", collapse = ",")

  # Set column names for stationary and non-stationary results tables
  eval(parse(text=paste0("names(RPestimates.stat) <- c(\"WaterYear\", \"AMAX\", \"WaterYearIndex\", \"ScaledWaterYearIndex\", ", namesRP, ")")))
  eval(parse(text=paste0("names(RPestimates.nonstat) <- c(\"WaterYear\", \"AMAX\", \"WaterYearIndex\", \"ScaledWaterYearIndex\", ", namesRP, ")")))

  # Plot estimates and confidence intervals for selected gauge and save as a .png file

  # Set plot title if the user has not provided one
  if(is.null(plotTitle)){
    title <- ifelse(nchar(gaugeName) > 0,
                    paste0("Gauging station ", gaugeNumber, " (", gaugeName, ") single site analysis \n Stationary and non-stationary (", tolower(distributionType), ") ", dist, " models (time covariate)"),
                    paste0("Gauging station ", gaugeNumber, " single site analysis \n Stationary and non-stationary (", tolower(distributionType), ") ", dist, " models (time covariate)"))
  }

  # Set file name
  pngName <- ifelse(nchar(gaugeName) > 0,
                    paste0("Outputs/Time covariate models/", gaugeNumber, " (", gaugeName, ") ", dist, " user-specified distribution results and confidence interval plots.png"),
                    paste0("Outputs/Time covariate models/", gaugeNumber, " ", dist, " user-specified distribution results and confidence interval plots.png"))

  png(pngName, width = 1500, height = 900, res=120)

  par(mfrow=c(1,2), oma = c(0, 0, 4, 0))

  np <- length(RPs)  # number of return periods
  ylab <- expression(paste("Flow (", m^3, "/s)"))  # set y label
  xlab <- "Water year"  # set x label

  # Plot return level plot
  par(mar=c(11, 4.6, 0, 0.6))
  y <- resultsTable$AMAX
  x <- resultsTable$WaterYear

  # Plot non-stationary results first
  # Find effective return levels (flows) for the required RPs by selecting columns containing 'estimate' in their name
  effrl.stat <- RPestimates.stat[, grep("estimate", names(RPestimates.stat))]  # effective return levels (stationary)
  effrl.nonstat <- RPestimates.nonstat[, grep("estimate", names(RPestimates.nonstat))]  # effective return levels (non-stationary)
  yl <- range(c(c(y), c(effrl.nonstat)), finite = TRUE)  # set y limits
  yl[2] <- max(max(RPestimates.nonstat[,-c(1,3,4)]), max(RPestimates.stat[,-c(1,3,4)]))  # set upper y limit to be the maximum value in the confidence intervals
  plot(x, y, type = "h", xlab = xlab, ylab = ylab, ylim = yl)  # plot the AMAX data
  for (i in 1:np) lines(x, effrl.nonstat[, i], lty = i+1, col = "#56B4E9", lwd = 2)
  # Plot the stationary return levels over the top
  for (i in 1:np) lines(x, effrl.stat[, i], lty = i+1, col = "#D55E00", lwd = 2)

  legend("bottomleft", inset = c(0,-0.48), legend = c("Annual maximum flows",
                                                      paste0(RPs, "-year (non-stationary)"),
                                                      paste0(RPs, "-year (stationary)")),
         lty = c(1,rep(2:(np+1),2)), col = c(1, rep("#56B4E9",np), rep("#D55E00",np)), lwd = c(1, rep(2,2*np)), ncol = 2, bg = "white", xpd = TRUE, cex = 0.85)

  # Plot the confidence intervals for one of the return periods
  par(mar=c(11, 5.1, 0, 0.8))

  # CIs
  effrl.stat.CI.RP <- RPestimates.stat[, grep(paste0("\\b", CI.RP,"-year"), names(RPestimates.stat))]  # effective return levels and CI for user specified RP (stationary)
  effrl.nonstat.CI.RP <- RPestimates.nonstat[, grep(paste0("\\b", CI.RP,"-year"), names(RPestimates.nonstat))]  # effective return levels and CI for user specified RP (non-stationary)

  plot(x, y, type = "h", xlab = xlab, ylab = ylab, ylim = yl)
  # Non-stationary
  lines(x, effrl.nonstat.CI.RP[, grep("estimate", names(effrl.nonstat.CI.RP))], lty = 1, col = "#56B4E9", lwd = 2)
  for (i in c(1,3)) lines(x, effrl.nonstat.CI.RP[, i], lty = 5, col = "#56B4E9")

  # Stationary
  lines(x, effrl.stat.CI.RP[, grep("estimate", names(effrl.stat.CI.RP))], lty = 1, col = "#D55E00", lwd = 2)
  for (i in c(1,3)) lines(x, effrl.stat.CI.RP[, i], lty = 5, col = "#D55E00")

  legend("bottomright", inset=c(0, -0.48), legend = c("Annual maximum flows",
                                                      paste0(CI.RP, "-year (non-stationary)"),
                                                      paste0(CI, "% confidence interval (non-stationary)"),
                                                      paste0(CI.RP, "-year (stationary)"),
                                                      paste0(CI, "% confidence interval (stationary)")),
         lty = c(1,rep(c(1,5),2)), col = c(1, rep("#56B4E9",2), rep("#D55E00",2)), lwd = c(1, rep(c(2,1),2)), ncol = 2, bg = "white", xpd = TRUE, cex = 0.85)

  # Add title
  mtext(title, outer = TRUE, cex = 1.2, line = 1.5)

  dev.off()
}

#' plot.all
#'
#' For a specified gauge, plot the stationary and three non-stationary GEV or GLO distributions
#' (with varying location, varying scale and varying location and scale parameters) from
#' using time as a covariate. The fit.time function must have been run first.
#'
#' @param folder string containing a path to a folder containing two sub-folders called 'Inputs' and 'Outputs'. The plots will be written out to the 'Outputs/Time covariate models' folder.
#' @param gauge string of the gauge number to plot. Must have been used in the fit.time function.
#' @param fitResults the output from the fitting function. This could either be from a saved location or may already be in the R environment if fit.time has just been run and the outputs saved to a named object.
#' @param RPs vector of return periods to plot.
#' @param plotTitle optional parameter. String containing title for output plot if you wish to change the automatically-generated title. This should only be used if plotting one return period, if the return period value is to be included in the title. The automatically-generated title contains the appropriate return period. To split over two lines use '\\n' where you want the line break to be.
#'
#' @return The outputs from the function are .png files (one per return period) containing plots showing results from all four fitted distributions.
#' @export
#'
#' @examples
#' folder <- "C:/Non-stationarity/Data"
#' gauge <- "24004"
#' fitResults <- readRDS("Outputs/Time covariate models/GEV fits from fit time function.RDS")
#' RPs <- c(2, 10, 20, 50, 100)
#' plot.all(folder, gauge, fitResults, RPs)
#' # Or set just one RP. In this case, it is also possible to set a plot title.
#' RPs <- 100
#' plotTitle <- "Single site analysis results for 24004 (100-year) using time as a covariate"
#' plot.all(folder, gauge, fitResults, RPs, plotTitle)
plot.all <- function(folder, gauge, fitResults, RPs, plotTitle = NULL){

  gaugeNumber <- gauge

  if(!is.null(plotTitle)) title <- plotTitle

  # Set the working directory
  setwd(folder)

  # Check for existence of Input and Output folders and stop and print warning if missing
  if(!file.exists("Inputs") | !file.exists("Outputs")) stop("Inputs and/or Outputs folder does not exist", call. = FALSE)
  # Check that gauge number is in fitResults
  if(!gaugeNumber %in% names(fitResults)) stop("Gauge number does not exist in fitResults", call. = FALSE)

  # Extract pre-processed gauge data from fit.time
  gaugeData <- eval(parse(text=paste0("fitResults$`", gaugeNumber, "`$data")))
  # Extract gauge name
  gaugeName <- gaugeData$GaugeName[1]
  # Extract distribution used in the fitting
  dist <- eval(parse(text=paste0("fitResults$`", gaugeNumber, "`$dist")))

  # Find maximum flow for all RPs for setting the y-axis limits
  # Read from results table from the output of the fit.time function
  results <- eval(parse(text=paste0("fitResults$`", gaugeNumber, "`$resultsTable")))
  AMAXcol <- grep("AMAX", colnames(results))  # which column contains the AMAX data
  resCols <- grep(paste0(RPs, "-year", collapse = "|"), colnames(results))  # which columns contain results
  RPresults <- results[, c(AMAXcol, resCols)]  # extract only the RPs to plot
  RPresults <- RPresults[grep("CI", colnames(RPresults), invert=TRUE)]  # remove CI columns
  maxFlow <- max(RPresults)

  # Set palette of colour-blind friendly colours (http://www.mikemeredith.net/blog/2018/Colours.htm)
  cbColours <- c( black = "#000000",         # black
                  vermilion = "#D55E00",     # red
                  green = "#009E73",         # green3
                  blue = "#0072B2",          # blue
                  sky = "#56B4E9",           # cyan
                  purple = "#CC79A7",        # magenta
                  yellow = "#F0E442",        # yellow
                  orange = "#E69F00")        # (grey)

  for (type in c("stat", "var.loc", "var.scale", "var.loc.scale")){

    fitCase <- paste0("fitResults$'", gaugeNumber, "'$fit.", type)

    # Find estimates and confidence intervals
    eval(parse(text=paste0(type, ".return.levels <- predict(", fitCase, ", M = RPs, ci.fit = TRUE, unique.=FALSE)")))

  }

  for(RP in RPs){
    # Compile results
    resultsTable <- eval(parse(text = paste0("data.frame(gaugeData, \"Stat\" = stat.return.levels$obj$M.", RP, "[,2],
                               \"Var.loc\" = var.loc.return.levels$obj$M.", RP, "[,2], \"Var.scale\" = var.scale.return.levels$obj$M.", RP, "[,2],
                               \"Var.loc.scale\" = var.loc.scale.return.levels$obj$M.", RP, "[,2])")))

    # Set plot title if the user has not provided one
    if(is.null(plotTitle)){
      title <- ifelse(nchar(gaugeName) > 0,
                      paste0("Gauging station ", gaugeNumber, " (", gaugeName, ") single site analysis \n All ", dist, " models (time covariate): ", RP, "-year"),
                      paste0("Gauging station ", gaugeNumber, " single site analysis \n All ", dist, " models (time covariate): ", RP, "-year"))
    }

    # Set file name
    pngName <- ifelse(nchar(gaugeName) > 0,
                      paste0("Outputs/Time covariate models/", gaugeNumber, " (", gaugeName, ") ", dist, " all distribution plots (", RP, "-year).png"),
                      paste0("Outputs/Time covariate models/", gaugeNumber, " ", dist, " all distribution plots (", RP, "-year).png"))

    # Plot results
    png(pngName, width = 1200, height = 1200, res = 120)

    ylab <- expression(paste("Flow (", m^3, "/s)"))  # set y label
    xlab <- "Water year"  # set x label

    par(mar=c(10, 4.6, 4.1, 2.1))
    x <- resultsTable$WaterYear
    y <- resultsTable$Flow
    yl <- range(c(y), finite = TRUE)  # set y limits
    yl[2] <- max(maxFlow, resultsTable$Flow)  # reset upper y limit

    plot(x, y, type = "h", xlab = xlab, ylab = ylab, ylim = yl, main = title, cex.lab = 1.2, cex.main = 1.2)  # plot the AMAX data
    types <- c("Stat", "Var.loc", "Var.scale", "Var.loc.scale")
    for (i in 1:4){
      type = types[i]
      eval(parse(text=paste0("lines(x, resultsTable$", type, ", lty = i+1, col = cbColours[i+1], lwd = 2)")))  # use i+1 so don't get solid black line which would stand out
    }

    legend("bottomleft", inset = c(0,-0.25), legend = c("Annual maximum flows", paste("Stationary", dist),
                                                        paste("Non-stationary (varying location)", dist),
                                                        paste("Non-stationary (varying scale)", dist),
                                                        paste("Non-stationary (varying location and scale)", dist)),
           lty = c(1:5), col = c(1, cbColours[2:5]), lwd = c(1, rep(2,4)), ncol = 2, bg = "white", xpd = TRUE, cex = 1)


    dev.off()

  }

}
