#' fit.phys.cov
#'
#' The fit.phys.cov function fits a stationary and multiple non-stationary GEV or GLO distributions to AMAX flow
#' data, using various combinations of physical covariates supplied by the user (a maximum of seven physical
#' covariates must be supplied). The user can then use the information returned about each model to decide which
#' model to use for estimating return levels (flows) for their required return periods.
#'
#' @param dist string of the distribution to use to fit models using physical covariates. Must be either "GEV" or "GLO".
#' @param folder string containing a path to a folder containing two sub-folders called 'Inputs' and 'Outputs'. 'Inputs' must contain one file per gauge containing annual maximum data. Data should be .AM files.
#' @param detrend.cov optional string. If included (i.e. not NULL), the other covariates are detrended based on the covariate in this string, usually time (represented by “ScaledWaterYearIndex” in this function). This is recommended for models where a physical covariate and time may both be included in the location or scale parameter (options 1 and 2 below). The default is “ScaledWaterYearIndex”. This can be changed to NULL if you do not wish to detrend the covariate data.
#' @param option optional numeric integer - must be 1, 2 (default) or 3. Option 1 fits models using all possible combinations of covariates, option 2 fits models using combinations which only include one physical covariate and/or time, and option 3 fits models using combinations which include one covariate at a time in the location and/or scale parameter.
#' @param best_fit_method optional string. Method of automatically selecting a best fit model that can be used in the next step if desired. Must be "AIC" or "BIC" (default).
#'
#' @return A list containing information for all gauges analysed is returned from the function and can be accessed if the outputs have been saved to a named object (as in the examples below). For each gauge, the following information is saved: dist (distribution used - taken from input argument), model_fits (described below), all_data_raw (the data used in the model fitting - not detrended), gaugeName (name of the gauge) and detrend.cov (the string from the input argument). Most of these are saved for use in the next step. The information is also automatically saved as a .RDS file for each gauge separately, for reference purposes only.
#' @return The model_fits element is another list containing the following information: best_model (details of the best fitting model selected based on AIC or BIC, as specified by the user), best_model_formula (details of the formula used to fit best_model), stat_model (details of the fitted stationary model) and all_models_metadata (a metadata file associated with fitting models to all combinations of covariates selected by the user). The latter is also written out as a .csv file.

#' @export
#'
#' @examples
#' folder <- "C:/Non-stationarity/Data"
#' dist <- "GEV"
#' # Run without detrending the covariate data
#' physCovModelFits <- fit.phys.cov(dist, folder, detrend.cov = NULL, option = 3, best_fit_method = "BIC")
#' # Run with detrending the covariate data based on the time covariate (i.e. "ScaledWaterYearIndex", the
#' # automatically-generated centred and scaled index of water year)
#' physCovModelFits <- fit.phys.cov(dist, folder, detrend.cov = "ScaledWaterYearIndex", option = 2, best_fit_method = "BIC")
fit.phys.cov <- function(dist, folder, detrend.cov = "ScaledWaterYearIndex", option = 2, best_fit_method = "BIC"){

  # Set working directory
  setwd(folder)

  # Set up output lists
  fits.phys <- failedInputs <- list()

  # Check for existence of Input and Output folders and stop and print warning if missing
  if(!file.exists("Inputs") | !file.exists("Outputs")) stop("Inputs and/or Outputs folder does not exist", call. = FALSE)

  # Check that distribution is either GEV or GLO
  if(! (dist == "GEV" || dist == "GLO")) stop("Distribution needs to be either GEV or GLO", call. = FALSE)

  # Set up folder for physical covariates outputs if one does not already exist
  if(!file.exists("Outputs/Physical covariates models")){
    dir.create("Outputs/Physical covariates models")
  }

  # List the files in the input folder
  gauges <- list.files(paste0(folder, "/Inputs"), pattern = ".AM", full.names = TRUE, ignore.case = TRUE)

  # Import look-up file to extract gauge name to go with the number
  # Suppress warnings and write out custom error message instead if the file cannot be accessed by R
  readTest <- suppressWarnings(try(metadata <- read.csv("Inputs/Gauge_metadata.csv", stringsAsFactors = FALSE), silent = TRUE))
  if(class(readTest) == "try-error"){
    stop("Error reading 'Gauge_metadata.csv' file. Please check that it exists in your 'Inputs' folder.", call. = FALSE)
  }

  #### ============================================== LOOP THROUGH GAUGES ============================================== ####
  # Add progress bar
  pb <- txtProgressBar(min = 0, max = length(gauges), style = 3)

  # Some info
  print(paste(length(gauges), "gauging stations queued for analysis:"))
  writeLines(gauges)
  writeLines(" ")

  myStartTime <- Sys.time()

  for (i in 1:length(gauges)){

    print(gauges[i])

    #### ============================================= DATA PRE-PROCESSING ============================================= ####
    preprocessedDataAll <- suppressWarnings(import.preprocess.data(gauges[i], metadata))
    preprocessedData <- preprocessedDataAll$res

    if(is.null(preprocessedDataAll$res)){
      failedInputs[[as.character(gauges[i])]] <- list(gauge = gauges[i], warn = preprocessedDataAll$warn, err = preprocessedDataAll$err)
      next
    }

    inData <- preprocessedData$inData
    gaugeNumber <- preprocessedData$gaugeNumber
    gaugeName <- inData$GaugeName[1]
    rejectedYears <- preprocessedData$rejectedYears

    # Skip gauge if covariates .csv file does not exist
    if(!file.exists(paste0("Inputs/", gaugeNumber, "_covariates.csv"))) next

    # Keep all years
    gaugeDataAll <- inData[c("GaugeNumber", "GaugeName", "FirstWaterYear", "LastWaterYear", "Missing", "Rejected", "WaterYear", "Date", "Flow")]  # extract just the columns that are required
    # Remove rejected years
    gaugeData <- gaugeDataAll[which(!gaugeDataAll$WaterYear %in% rejectedYears),]
    # Remove rows where flow is missing
    gaugeData <- gaugeData[which(!is.na(gaugeData$Flow)),]
    gaugeData$NumberOfYears <- length(gaugeData$WaterYear)  # number actually used (excluding rejected years)

    # Read in covariate data
    covariates <- fread(paste0("Inputs/", gaugeNumber, "_covariates.csv"), header = T)

    # Join data by common year
    all_data_raw <- left_join(gaugeData, covariates, by = c("WaterYear" = "Year"))
    # If the covariate data file contains fewer years than the AMAX data, the covariate values will be NA for the additional AMAX years.

    # Set scaled water year index covariate - representing time
    all_data_raw$WaterYearIndex <- gaugeData$WaterYear - gaugeData$WaterYear[1]  # start the index column at the beginning of the subsetted data
    all_data_raw$ScaledWaterYearIndex <- scale(all_data_raw$WaterYearIndex) # scales to mean = 0, sd = 1

    # Remove NAs keeping all covariates
    FlowCol <- all_data_raw$Flow
    colsToRemove <- c("GaugeNumber", "GaugeName", "FirstWaterYear", "LastWaterYear", "Missing", "Rejected", "Date", "NumberOfYears", "Flow")
    all_data_raw <- all_data_raw[, !names(all_data_raw) %in% colsToRemove]
    all_data_raw <- data.frame(Flow = FlowCol, all_data_raw)  # create data frame with Flow in the first column
    all_data_raw <- na.omit(all_data_raw)

    # Check that a maximum of 7 physical covariates have been provided and write an error message if there are too many
    if(ncol(all_data_raw)>11) stop("You have too many physical covariates. Please check that you have a maximum of seven physical covariates in your input data.", call. = FALSE)

    # Run the BestModel function (internal) - these outputs will be needed for the next step
    model_fits <- BestModel(dist, all_data_raw, detrend.cov = detrend.cov, option = option, best_fit_method = best_fit_method)

    # Find if there are any warnings from the fitting and write out the metadata associated with the fitting
    out_metadata <- model_fits$all_models_metadata

    # Suppress warnings and write out custom error message instead if the file cannot be accessed by R
    outPath <- ifelse(nchar(gaugeName) > 0, paste0("Outputs/Physical covariates models/" , gaugeNumber, " (", gaugeName, ") ", dist, " fitting information (physical covariates).csv"),
                      paste0("Outputs/Physical covariates models/" , gaugeNumber, " ", dist, " fitting information (physical covariates).csv"))
    writeTest <- suppressWarnings(try(write.csv(out_metadata, file = outPath), silent = TRUE))
    if(class(writeTest) == "try-error"){
      stop(paste0("Error saving physical covariates fitting information file for ", gaugeNumber, ". Please check that you do not have the file open."), call. = FALSE)
    }

    outPath <- ifelse(nchar(gaugeName) > 0, paste0("Outputs/Physical covariates models/", gaugeNumber, " (", gaugeName, ") ", dist, " model fits (physical covariates).RDS"),
                      paste0("Outputs/Physical covariates models/", gaugeNumber, " ", dist, " model fits (physical covariates).RDS"))
    saveRDS(list(dist = dist, model_fits = model_fits, all_data_raw = all_data_raw, gaugeName = gaugeName, detrend.cov = detrend.cov), file = outPath)

    # Save fit for each gauge
    eval(parse(text=paste0("fits.phys.", gaugeNumber, " <- list(dist = dist, model_fits = model_fits, all_data_raw = all_data_raw, gaugeName = gaugeName, detrend.cov = detrend.cov)")))
    # Add to list
    fits.phys[[as.character(gaugeNumber)]] <- eval(parse(text=paste0("fits.phys.", gaugeNumber)))

    Sys.sleep(0.1)
    setTxtProgressBar(pb, i) # updates progress bar after every gauge has been processed
  }

  # Tidy up for loop processing
  close(pb)
  print("*** Total processing time:")
  print(Sys.time()-myStartTime)

  #### =================================== WRITE OUT INPUT DATA WARNINGS AND ERRORS =================================== ####
  #### ========================================== OUTPUT FILE HEADERS TO CSV ========================================== ####
  # Set up header row
  # Only run this once so don't overwrite
  inputWEHeader <- c("File", "Warnings", "Errors")

  # Suppress warnings and write out custom error message instead if the file cannot be accessed by R
  writeTest <- suppressWarnings(try(write(inputWEHeader, "Outputs/Physical covariates models/Input data warnings and errors.csv", ncolumns = length(inputWEHeader), sep = ","), silent = TRUE))
  if(class(writeTest) == "try-error"){
    stop("Error saving input data warnings and errors file. Please check that you do not have the file open.", call. = FALSE)
  }

  # This will only capture the first error as it stops after the first one
  for (file in names(failedInputs)){
    failInfo <- eval(parse(text=paste0("failedInputs$`", as.character(file), "`")))
    Warn <- paste0(failInfo$warn, collapse = " & ")
    Err <- paste0(failInfo$err, collapse = " & ")
    inputWEInfo <- c(failInfo$gauge, Warn, Err)
    # Account for commas in the warning and error messages so that they are not split across columns
    inputWEInfo2 <- paste("\"", inputWEInfo, "\"", sep = "", collapse = ",")
    write(inputWEInfo2, "Outputs/Physical covariates models/Input data warnings and errors.csv", ncolumns = length(inputWEHeader), sep = ",", append = T)
  }

  return(fits.phys)
}


#' res.phys.cov
#'
#' The res.phys.cov function uses one of the fitted non-stationary models from the 'fit.phys.cov' function
#' to estimate conditional and marginal return levels (and present-day marginal return levels if required).
#' Stationary return levels are estimated based on the fitted stationary model.
#' This function also outputs plots of the results.
#'
#' @param folder string containing a path to a folder containing two sub-folders called 'Inputs' and 'Outputs'. 'Inputs' must contain one file per gauge containing annual maximum data. Data should be .AM files.
#' @param physCovModelFits output from the fit.phys.cov function
#' @param gauge string of the number of the gauge to analyse. Must have been used in the fit.phys.cov function and must match the number in the fitting information .csv file output from the fit.phys.cov function (so if the number has been edited then it should be the edited number, e.g. 24004x rather than 24004).
#' @param RPs numeric vector of return periods for which frequency analysis is required. A maximum of eight return periods can be plotted in the 'model results' plot - if more are listed here then only the first eight will be selected for the plot. All return periods will be used for the frequency analysis and for the 'encounter probability' plot.
#' @param CI numeric value. Confidence interval to be calculated for results as a percentage, e.g. ‘90’ for 90\% confidence interval is the default.
#' @param model_type string of "Best" or "User". Select whether to use the best fit model based on AIC/BIC or a user-selected model to calculate return levels (and then plot).
#' @param locFit string only required if model_type = "User". locFit is the regression equation of covariates to include in the location parameter (copy from 'fitting information (physical covariates)' .csv file output by the fit.phys.cov function), e.g. locFit = "~ 1 + NAO_SON + ScaledWaterYearIndex".
#' @param scaleFit string only required if model_type = "User". scaleFit is the regression equation of covariates to include in the scale parameter (copy from 'fitting information (physical covariates)' .csv file output by the fit.phys.cov function), e.g. scaleFit = "~ 1 + ScaledWaterYearIndex".
#' @param present_day logical - set to TRUE to calculate the present-day marginal return level as well as the marginal return level, otherwise set to FALSE. This can only be TRUE if time ("ScaledWaterYearIndex") is included as a covariate in the selected model.
#'
#' @return A .RDS file is returned which contains information relating to the analysis for the gauge (from the model used to estimate return levels, to the results of the frequency analysis). This can be opened in R and referred to at a later date if required. Much of the information is also written out in .csv files.
#' @return A range of .csv files are also returned which contain information about the formula of the non-stationary model used, to the conditional and marginal return levels (plus present-day marginal return levels, if selected) from the non-stationary model and the return levels from the stationary model.
#' @return Three plots are also written out as .png files. The first ('correlation plot') is of flow plotted against all covariates (detrended if this option was selected). The second ('model results plot') is of flow plotted against conditional return levels from the non-stationary model and stationary return levels from the stationary model for the observed period. The third ('encounter probability plot') is an encounter probability plot showing the marginal and stationary results.
#' @export
#'
#' @examples
#' folder <- "C:/Non-stationarity/Data"
#' physCovModelFits <- physCovModelFits  # output from running the fit.phys.cov function
#' gauge <- "24004"
#' RPs <- c(2, 10, 20, 50, 100)
#' CI <- 90
#' # Using the best model based on AIC or BIC (as selected in fit.phys.cov)
#' res.phys.cov(folder, physCovModelFits, gauge, RPs, CI, model_type = "Best", locFit = NULL, scaleFit = NULL, present_day = TRUE)
#' # Using a user-selected model
#' res.phys.cov(folder, physCovModelFits, gauge, RPs, CI, model_type = "User", locFit = "~ 1 + Rain_WY_mean + ScaledWaterYearIndex", scaleFit = "~ 1 + ScaledWaterYearIndex", present_day = TRUE)
res.phys.cov <- function(folder, physCovModelFits, gauge, RPs, CI = 90, model_type = "Best", locFit = NULL, scaleFit = NULL, present_day){

  setwd(folder)

  gaugeNumber <- gauge

  # Extract information from physCovModelFits
  dist = eval(parse(text=paste0("physCovModelFits$`", gaugeNumber, "`$dist")))
  model_fits = eval(parse(text=paste0("physCovModelFits$`", gaugeNumber, "`$model_fits")))
  all_data_raw = eval(parse(text=paste0("physCovModelFits$`", gaugeNumber, "`$all_data_raw")))  # raw data, not detrended
  gaugeName = eval(parse(text=paste0("physCovModelFits$`", gaugeNumber, "`$gaugeName")))
  detrend.cov = eval(parse(text=paste0("physCovModelFits$`", gaugeNumber, "`$detrend.cov")))

  # Check that present_day equals TRUE or FALSE and stop and print warning if not
  if(!(present_day == TRUE | present_day == FALSE)) stop("present_day must be TRUE or FALSE", call. = FALSE)

  # Check that model_type equals "Best" or "User" and stop and print warning if not
  if(!(model_type == "Best" | model_type == "User")) stop("model_type must be \"Best\" or \"User\"", call. = FALSE)

  # Check that location and scale parameters have not been entered if model_type = "Best"
  if(model_type == "Best" & (!is.null(locFit) | !is.null(scaleFit))) stop("model_type is selected as \"Best\" but locFit and/or scaleFit is not NULL", call. = FALSE)

  # If user-selected model is to be used, i.e. model_type = "User", then check that
  # location and scale parameters have been provided and fit the model
  if(model_type == "User"){
    print("model_type is User")
    if(is.null(locFit) | is.null(scaleFit)) stop("model_type is selected as \"User\" but locFit and/or scaleFit is NULL", call. = FALSE)
    userModel <- UserModel(dist, all_data_raw, locFit, scaleFit, detrend.cov = detrend.cov)
  } else{
    userModel = NULL
  }

  fit_results <- CovariateModel(dist, all_data_raw, ret_periods = RPs, CI, model_fits = model_fits, model_type = model_type, userModel = userModel,
                                locFit = locFit, scaleFit = scaleFit, gaugeNumber, gaugeName, detrend.cov, present_day)

  # Set up save directory
  savePath <- ifelse(nchar(gaugeName) > 0, paste0("Outputs/Physical covariates models/", gaugeNumber, " (", gaugeName, ") ", dist),
                     paste0("Outputs/Physical covariates models/", gaugeNumber, " ", dist))

  # Write out everything
  saveRDS(fit_results, file = paste0(savePath, " model.RDS"))

  # Access information about the fitted models
  model <- fit_results$model
  statModel <- fit_results$stat_model

  # Write out the model formula
  modelFormula <- fit_results$model_formula
  modelFormulaTable <- data.frame(location = modelFormula[1], scale = modelFormula[2])
  write.csv(modelFormulaTable, file = paste0(savePath, " model formula.csv"), row.names = FALSE)

  # Write out the conditional return levels with confidence intervals
  condRLs <- fit_results$condRLs
  if(ncol(condRLs) == 1){  # if stationary model and hence no conditional return levels have been calculated
    write.table(condRLs, file = paste0(savePath, " conditional return levels.csv"), sep = ",", row.names = FALSE, col.names = FALSE)
  } else{
    write.csv(condRLs, file = paste0(savePath, " conditional return levels.csv"), row.names = FALSE)
  }

  # Write out the stationary return levels with confidence intervals
  statRLs <- fit_results$statRLs
  write.csv(statRLs, file = paste0(savePath, " stationary return levels.csv"), row.names = FALSE)

  # Write out the marginal return levels
  margRLs <- fit_results$margRLs
  if(ncol(margRLs) == 1){  # if stationary model and hence no marginal return levels have been calculated
    write.table(margRLs, file = paste0(savePath, " marginal return levels.csv"), sep = ",", row.names = FALSE, col.names = FALSE)
  } else{
    write.csv(margRLs, file = paste0(savePath, " marginal return levels.csv"), row.names = FALSE)
  }

  # Write out present-day marginal return levels if appropriate
  pres_margRLs <- fit_results$pres_margRLs
  if(!is.null(pres_margRLs)){
    write.csv(pres_margRLs, file = paste0(savePath, " present-day marginal return levels.csv"), row.names = FALSE)
  }

  # Access information about the marginal bootstrapping to check on warnings and errors

  # Number of warnings in the fitted bootstrap models
  boot_model_warnings <- fit_results$margRLsAllInfo$boot_model_warnings
  # Extract number of warnings
  numWarn <- length(boot_model_warnings[boot_model_warnings != 0])
  # Number of bootstrap samples has been set to 200
  # Print message if required
  if(numWarn != 0) print(paste0(numWarn, " bootstrapping sample/s out of 200 produced warnings when fitting the model"))

  # Number of errors in the fitted bootstrap models
  boot_model_errors <- fit_results$margRLsAllInfo$boot_model_errors
  # Extract number of errors
  numErr <- length(boot_model_errors[boot_model_errors != 0])
  # Print message if required
  if(numErr != 0) print(paste0(numErr, " bootstrapping sample/s out of 200 produced errors when fitting the model"))

  # Warnings from calculating marginal return levels from the bootstrap models
  boot_marg_RLs_warnings <- fit_results$margRLsAllInfo$boot_marg_RLs_warnings
  # Extract number of warnings
  numWarnRLs <- sum(!is.na(boot_marg_RLs_warnings))
  # Print message if required
  if(numWarnRLs != 0) print(paste0("There was/were ", numWarnRLs, " warning/s from calculating marginal return levels from 200 bootstrap models for calculating the confidence intervals over the ", length(RPs), " return periods."))

  # Errors from calculating marginal return levels from the bootstrap models
  boot_marg_RLs_errors <- fit_results$margRLsAllInfo$boot_marg_RLs_errors
  # Extract number of errors
  numErrRLs <- sum(!is.na(boot_marg_RLs_errors))
  # Print message if required
  if(numErrRLs != 0) print(paste0("There was/were ", numErrRLs, " error/s from calculating marginal return levels from 200 bootstrap models for calculating the confidence intervals over the ", length(RPs), " return periods."))

  # Warnings from calculating present-day marginal RLs from the bootstrap models
  boot_pres_marg_RLs_warnings <- fit_results$margRLsAllInfo$boot_pres_marg_RLs_warnings
  # Extract number of warnings
  numWarnRLs_pres <- sum(!is.na(boot_pres_marg_RLs_warnings))
  # Print message if required
  if(numWarnRLs_pres != 0) print(paste0("There was/were ", numWarnRLs_pres, " warning/s from calculating present-day marginal return levels for 200 bootstrap models for calculating the confidence intervals over the ", length(RPs), " return periods."))

  # Errors from calculating present-day marginal RLs from the bootstrap models
  boot_pres_marg_RLs_errors <- fit_results$margRLsAllInfo$boot_pres_marg_RLs_errors
  # Extract number of errors
  numErrRLs_pres <- sum(!is.na(boot_pres_marg_RLs_errors))
  # Print message if required
  if(numErrRLs_pres != 0) print(paste0("There was/were ", numErrRLs_pres, " error/s from calculating present-day marginal return levels for 200 bootstrap models for calculating the confidence intervals over the ", length(RPs), " return periods."))

  # Plot the results
  plots <- PlotFit(dist, fit = fit_results, ret_periods = RPs, all_data_raw, CI, gaugeNumber, gaugeName, detrend.cov)

  # Save plots
  pngName <- paste0(savePath, " correlation plot.png")
  png(pngName, width = 1200, height = 1200, res = 120)
  print(plots$f)
  dev.off()

  pngName <- paste0(savePath, " model results plot.png")
  png(pngName, width = 1200, height = 1200, res = 120)
  print(plots$g)
  dev.off()

  pngName <- paste0(savePath, " encounter probability plot.png")
  png(pngName, width = 1200, height = 1200, res = 120)
  print(plots$h)
  dev.off()

}
