#### =========================================== MANN-KENDALL INTERNAL FUNCTIONS =========================================== ####
#' MKtest
#'
#' MKtest is an internal function used by MK.test. Carries out the Mann-Kendall trend test (using the trend package).
#' @keywords internal
MKtest <- function(gaugeData, sigLevel){

  # Convert significance level from a percentage to a decimal number
  sigValue <- sigLevel/100

  MKresults <- mk.test(gaugeData$Flow)

  p_value <- MKresults$p.value

  # Interpretation
  Interpretation <- ifelse(p_value < sigValue, paste0("Reject H0. Statistically significant trend at ", sigLevel, "% significance level."),
                           paste0("No statistically significant trend at ", sigLevel, "% significance level."))

  # Direction of trend
  if(p_value < sigValue) {
    trendDirection <- ifelse(sign(MKresults$estimates[3]) == -1, "Negative", "Positive")
  } else {
    trendDirection <- "N/A"
    }

  # Combine results
  MKoutputs <- c(gaugeData$GaugeNumber[1], gaugeData$GaugeName[1], gaugeData$FirstWaterYear[1], gaugeData$LastWaterYear[1],
                 gaugeData$Missing[1], gaugeData$Rejected[1], gaugeData$NumberOfYears[1], MKresults$estimates[1], MKresults$estimates[2],
                 MKresults$estimates[3], MKresults$statistic, p_value, Interpretation, trendDirection, format(Sys.time(), "%d/%m/%Y %X"))

  # Write out results to .csv file
  write(MKoutputs, "Outputs/Mann-Kendall/Mann-Kendall results.csv", ncolumns = length(MKoutputs), sep = ",", append = T)

}

#' trendPlots
#'
#' Internal function to create plots to accompany the Mann-Kendall testing.
#' @keywords internal
trendPlots <- function(gaugeData){
  outPath <- ifelse(nchar(gaugeData$GaugeName[1]) > 0,
                    paste0("Outputs/Mann-Kendall/", gaugeData$GaugeNumber[1], " (", gaugeData$GaugeName[1], ") trend plots.png"),
                    paste0("Outputs/Mann-Kendall/", gaugeData$GaugeNumber[1], " trend plots.png"))
  pngName <- outPath
  png(pngName, width = 900, height = 750)
  par(mfrow = c(2,1))  # set plot window to 2 rows and 1 column
  par(mar = c(5.1, 4.6, 4.1, 10.5))
  # Set title string over two lines
  firstLine <- ifelse(nchar(gaugeData$GaugeName[1]) > 0,
                      paste0(as.character(gaugeData$GaugeNumber[1]), " (", gaugeData$GaugeName[1], ")"),
                      paste0(as.character(gaugeData$GaugeNumber[1])))
  titleLines <-
    list(firstLine,
         "",
         paste0("AMAX flow data"))
  # Plot AMAX data with smoothed line
  plot(gaugeData$WaterYear, gaugeData$Flow, type = "h", main = paste(titleLines), xlab = "Water year",
       ylab = expression(paste("Flow (", m^3/s, ")")), cex.axis = 1.2, cex.lab = 1.2, cex.main = 1.3)
  lines(lowess(gaugeData$WaterYear,gaugeData$Flow), lwd=3, col=2)
  legend("right", legend = c("Observed", "Smoothed\n(LOWESS)"), col = c(1,2), lty = c(1,1), lwd=c(1,3), inset = c(-0.2, 0), xpd = TRUE, bty = "n", cex = 1.2)
  # Plot ACF
  acf(gaugeData$Flow, main = NA, cex.axis = 1.2, cex.lab = 1.2)  # computes estimates of the autocorrelation function
  mtext("Autocorrelation", side=3, line=1, outer=F, cex=1.3, font=2)
  legend("right", legend = c("Autocorrelation\nestimate", "95%\nconfidence\ninterval"), col = c(1,4), lty = c(1,2), inset = c(-0.22, 0), xpd = TRUE, bty = "n", cex = 1.2)
  dev.off()
}

#### ========================================== MANN-KENDALL TREND TEST FUNCTION ========================================== ####
#' MK.test
#'
#' MK.test carries out the Mann-Kendall trend test on all gauges listed in a folder. Results are automatically written out
#' to the 'Outputs/Mann-Kendall' folder. The 'Mann-Kendall' folder will be automatically generated if it does not already exist.
#'
#' @param folder string containing a path to a folder containing two sub-folders called 'Inputs' and 'Outputs'. 'Inputs' must contain one file per gauge containing annual maximum data. Data should be .AM files.
#' @param sig.level numeric value giving significance level as a percentage for interpretation of results in the output .csv file (e.g. input '5' to get an interpretation of whether trend is significant at the 5\% level). 5 is the default setting for this parameter.
#'
#' @return A .csv file containing results is written out. An explanation of the results can be found in the nonstat package user guide and in the trend package vignette.
#' @return One plot per gauge containing two graphs is also automatically written out. The first shows a time series of the AMAX data and smoothed data, using locally-weighted polynomial regression (LOWESS). The second shows the autocorrelation function.
#' @export
#'
#' @examples
#' # Set the inputs to the function
#' folder <- "C:/Non-stationarity/Data"  # this folder contains sub-folders called 'Inputs' and 'Outputs'
#' sig.level <- 5  # set significance level to be 5% for interpretation purposes only
#' # Run the function
#' MK.test(folder, sig.level)
MK.test <- function(folder, sig.level = 5){

  # Set working directory to be the user-specified folder
  setwd(folder)

  # Check for existence of Input and Output folders and stop and print warning if missing
  if(!file.exists("Inputs") | !file.exists("Outputs")) stop("Inputs and/or Outputs folder does not exist", call. = FALSE)

  # Set up folder for MK outputs if one does not already exist
  if(!file.exists("Outputs/Mann-Kendall")){
    dir.create("Outputs/Mann-Kendall")
  }

  # List the files in the input folder
  gauges <- list.files(paste0(folder, "/Inputs"), pattern = ".AM", full.names = TRUE, ignore.case = TRUE)

  # Import look-up file to extract gauge name to go with the number
  # Suppress warnings and write out custom error message instead if the file cannot be accessed by R
  readTest <- suppressWarnings(try(metadata <- read.csv("Inputs/Gauge_metadata.csv", stringsAsFactors = FALSE), silent = TRUE))
  if(class(readTest) == "try-error"){
    stop("Error reading 'Gauge_metadata.csv' file. Please check that it exists in your 'Inputs' folder.", call. = FALSE)
  }

  # Set up list output for gauges that fail during pre-processing
  failedInputs <- list()

  #### ======================================= OUTPUT FILE HEADERS TO CSV ======================================= ####
  # Only run this once so don't overwrite
  header <- c("Gauge Number", "Gauge Name", "First Water Year", "Last Water Year", "Missing Years", "Rejected Years",
              "Number Of Years", "Score (S)", "Var(Score)", "Tau", "Z score", "2-sided p-value",
              "Interpretation", "Trend direction", "Timestamp")
  # Suppress warnings and write out custom error message instead if the file cannot be accessed by R
  writeTest <- suppressWarnings(try(write(header, "Outputs/Mann-Kendall/Mann-Kendall results.csv", ncolumns = length(header), sep = ","), silent = TRUE))
  if(class(writeTest) == "try-error"){
    stop("Error saving Mann-Kendall results file. Please check that you do not have the file open.", call. = FALSE)
  }

  #### =========================================== LOOP THROUGH GAUGES =========================================== ####
  # Add progress bar
  pb <- txtProgressBar(min = 0, max = length(gauges), style = 3)

  # Some info
  print(paste(length(gauges), "gauging stations queued for analysis:"))
  writeLines(gauges)
  writeLines(" ")

  myStartTime <- Sys.time()

  for (i in 1:length(gauges)){
    print(gauges[i])

    #### ========================================== DATA PRE-PROCESSING ========================================== ####
    preprocessedDataAll <- suppressWarnings(import.preprocess.data(gauges[i], metadata))
    preprocessedData <- preprocessedDataAll$res

    if(is.null(preprocessedDataAll$res)){
      failedInputs[[as.character(gauges[i])]] <- list(gauge = gauges[i], warn = preprocessedDataAll$warn, err = preprocessedDataAll$err)
      next
    }

    inData <- preprocessedData$inData
    gaugeNumber <- preprocessedData$gaugeNumber
    rejectedYears <- preprocessedData$rejectedYears

    # Extract just the columns that are required (all years)
    gaugeDataAll <- inData[c("GaugeNumber", "GaugeName", "FirstWaterYear", "LastWaterYear", "Missing", "Rejected", "WaterYear", "Date", "Flow")]
    # Remove rejected years
    gaugeData <- gaugeDataAll[which(!gaugeDataAll$WaterYear %in% rejectedYears),]
    # Remove rows where flow is missing
    gaugeData <- gaugeData[which(!is.na(gaugeData$Flow)),]
    gaugeData$NumberOfYears <- length(gaugeData$WaterYear)  # number actually used (excluding rejected years)

    #### ===================================== APPLY MANN-KENDALL TREND TEST ===================================== ####
    trendPlots(gaugeData)
    MKtest(gaugeData, sig.level)

    Sys.sleep(0.1)
    setTxtProgressBar(pb, i) # updates progress bar after every gauge has been processed
  }

  #### ======================================== INPUT DATA WARNINGS AND ERRORS ======================================== ####
  #### ========================================== OUTPUT FILE HEADERS TO CSV ========================================== ####
  # Set up header row
  # Only run this once so don't overwrite
  inputWEHeader <- c("File", "Warnings", "Errors")

  # Suppress warnings and write out custom error message instead if the file cannot be accessed by R
  writeTest <- suppressWarnings(try(write(inputWEHeader, "Outputs/Mann-Kendall/Input data warnings and errors.csv", ncolumns = length(inputWEHeader), sep = ","), silent = TRUE))
  if(class(writeTest) == "try-error"){
    stop("Error saving input data warnings and errors file. Please check that you do not have the file open.", call. = FALSE)
  }

  #### ========================================= WRITE OUT WARNINGS AND ERRORS ========================================= ####
  # This will only capture the first error as it stops after the first one
  for (file in names(failedInputs)){
    failInfo <- eval(parse(text=paste0("failedInputs$`", as.character(file), "`")))
    Warn <- paste0(failInfo$warn, collapse = " & ")
    Err <- paste0(failInfo$err, collapse = " & ")
    inputWEInfo <- c(failInfo$gauge, Warn, Err)
    # Account for commas in the warning and error messages so that they are not split across columns
    inputWEInfo2 <- paste("\"", inputWEInfo, "\"", sep = "", collapse = ",")
    write(inputWEInfo2, "Outputs/Mann-Kendall/Input data warnings and errors.csv", ncolumns = length(inputWEHeader), sep = ",", append = T)
  }

  ## Tidy up for loop processing
  close(pb)
  print("*** Total processing time:")
  print(Sys.time()-myStartTime)

}

#### =============================================== PELT INTERNAL FUNCTION =============================================== ####
#' PELTtest
#'
#' PELTtest is an internal function used by PELT.test to carry out the PELT test (using the changepoint package).
#' @keywords internal
PELTtest <- function(gaugeData){

  peltResults <- list()

  # Apply log-transform to flow to make it approximately normal
  gaugeData$LogFlow <- log(gaugeData$Flow)

  # Normal Q-Q plot of the log-transformed data
  normQQoutPath <- ifelse(nchar(gaugeData$GaugeName[1]) > 0, paste0("Outputs/PELT/", gaugeData$GaugeNumber[1], " (", gaugeData$GaugeName[1], ") Normal Q-Q plot.png"),
                    paste0("Outputs/PELT/", gaugeData$GaugeNumber[1], " Normal Q-Q plot.png"))

  normqqPlotTitle <- ifelse(nchar(gaugeData$GaugeName[1]) > 0, paste0(gaugeData$GaugeNumber[1], " (", gaugeData$GaugeName[1], ") \n Normal Q-Q plot"),
                            paste0(gaugeData$GaugeNumber[1], " \n Normal Q-Q plot"))

  png(filename = normQQoutPath)
  stats::qqnorm(gaugeData$LogFlow, main = normqqPlotTitle)
  qqline(gaugeData$LogFlow)
  dev.off()

  # Apply PELT with minimum segment length (default penalty)
  cptModel = cpt.meanvar(gaugeData$LogFlow, method = "PELT", test.stat = "Normal", minseglen = 10)

  # Extract change points and number of changes
  cpt <- cptModel@cpts
  numChanges <- length(cptModel@cpts) - 1  # remove last one as this is the end of the record

  # Summary statistics (max of 2 change points for 10-year minimum segment length)
  mean_change <- mean_percent_change <- sd_change <- sd_percent_change <- mean_trend_direction <- year_change <-
    sd_trend_direction <- rep(NA,2)

  # If a change point is detected, calculate the mean before and after the change and the difference (absolute and percentage)
  if(numChanges > 0) {
    for (j in 1:length(cpt[-length(cpt)])) {
      start = ifelse(j==1, 1, cpt[j-1])
      year_change[j] <- gaugeData$WaterYear[cpt[j]]
      mean_change[j] <- mean(gaugeData$Flow[cpt[j]:cpt[j+1]]) - mean(gaugeData$Flow[start:(cpt[j]-1)])
      mean_percent_change[j] <- (mean_change[j]/mean(gaugeData$Flow[start:(cpt[j]-1)]))*100
      sd_change[j] <- sd(gaugeData$Flow[cpt[j]:cpt[j+1]]) - sd(gaugeData$Flow[start:(cpt[j]-1)])
      sd_percent_change[j] <- (sd_change[j]/sd(gaugeData$Flow[start:(cpt[j]-1)]))*100
      mean_trend_direction[j] <- ifelse(mean_change[j] < 0, "Negative", "Positive")
      sd_trend_direction[j] <- ifelse(sd_change[j] < 0, "Negative", "Positive")
    }
  }

  # Plot time series segmentation
  plotTitle <- ifelse(nchar(gaugeData$GaugeName[1]) > 0,
                      paste0(gaugeData$GaugeNumber[1], " (", gaugeData$GaugeName[1], ")\nPELT test"),
                      paste0(gaugeData$GaugeNumber[1], "\nPELT test"))

  if(all(is.na(year_change))) {
    g <- ggplot() + geom_line(aes(x = WaterYear, y = Flow, colour = "AMAX"), data = gaugeData, na.rm = T) + theme_bw() + ggtitle(plotTitle) +
      theme(plot.title = element_text(hjust = 0.5)) + ylab(expression(paste("Flow ", (m^3/s)))) + xlab("Water year") +
      scale_colour_manual("", values = c("AMAX" = "black"))
  } else {
    g <- ggplot() + geom_line(aes(x = WaterYear, y = Flow, colour="AMAX"), data = gaugeData, na.rm = T) +
      geom_vline(data = data.frame(year_change = year_change), aes(xintercept = year_change, colour = "Change point"), na.rm = T, show.legend = FALSE) +
      ggtitle(plotTitle) + theme_bw() +
      theme(plot.title = element_text(hjust = 0.5)) + ylab(expression(paste("Flow ", (m^3/s)))) + xlab("Water year") +
      scale_color_manual("", values = c("AMAX" = "black", "Change point" = "red"))
  }

  outPath <- ifelse(nchar(gaugeData$GaugeName[1]) > 0, paste0("Outputs/PELT/", gaugeData$GaugeNumber[1], " (", gaugeData$GaugeName[1], ") PELT results.png"),
                    paste0("Outputs/PELT/", gaugeData$GaugeNumber[1], " PELT results.png"))
  ggsave(filename = outPath, g)

  # Combine results
  peltResults <- c(gaugeData$GaugeNumber[1], gaugeData$GaugeName[1], length(gaugeData$WaterYear), numChanges, year_change,
                        mean_trend_direction, mean_change, mean_percent_change,
                        sd_trend_direction, sd_change, sd_percent_change, format(Sys.time(), "%d/%m/%Y %X"))

  return(peltResults)

}

#### =============================================== PELT TEST FUNCTION =============================================== ####
#' PELT.test
#'
#' PELT.test carries out change point detection using the Pruned Exact Linear Time (PELT) method on all gauges listed in a folder. Results are automatically written out
#' to the 'Outputs/PELT' folder. The 'PELT' folder will be automatically generated if it does not already exist.
#'
#' @param folder string containing a path to a folder containing two sub-folders called 'Inputs' and 'Outputs'. 'Inputs' must contain one file per gauge containing annual maximum data. Data should be .AM files.
#'
#' @return A .csv file is automatically generated containing the results for all gauges.
#' @return A plot is also generated for each gauge showing the AMAX data with any detected change point shown as a vertical line (some gauges will have no detected change point).
#' @return A Normal Q-Q plot is also generated to help determine if the assumption that the data follow a log-Normal distribution is reasonable.
#' @export
#'
#' @examples
#' # Set the inputs to the function
#' folder <- "C:/Non-stationarity/Data"  # this folder contains sub-folders called 'Inputs' and 'Outputs'
#' # Run the function
#' PELT.test(folder)
PELT.test <- function(folder){
  # Set working directory to be the user-specified folder
  setwd(folder)

  # Check for existence of Input and Output folders and stop and print warning if missing
  if(!file.exists("Inputs") | !file.exists("Outputs")) stop("Inputs and/or Outputs folder does not exist", call. = FALSE)

  # Set up folder for PELT outputs if one does not already exist
  if(!file.exists("Outputs/PELT")){
    dir.create("Outputs/PELT")
  }

  # List the files in the input folder
  gauges <- list.files(paste0(folder, "/Inputs"), pattern = ".AM", full.names = TRUE, ignore.case = TRUE)

  # Import look-up file to extract gauge name to go with the number
  # Suppress warnings and write out custom error message instead if the file cannot be accessed by R
  readTest <- suppressWarnings(try(metadata <- read.csv("Inputs/Gauge_metadata.csv", stringsAsFactors = FALSE), silent = TRUE))
  if(class(readTest) == "try-error"){
    stop("Error reading 'Gauge_metadata.csv' file. Please check that it exists in your 'Inputs' folder.", call. = FALSE)
  }

  # Set up list output for gauges that fail during pre-processing
  failedInputs <- list()

  # Set up output for listing gauges that have a record length that is too short
  writeTest <- suppressWarnings(try(write(c("Gauge number", "Gauge name", "Record length (years)"), "Outputs/PELT/Gauges with record lengths too short to be analysed.csv", ncolumns = 3, sep = ","), silent = TRUE))
  if(class(writeTest) == "try-error"){
    stop("Error saving 'Gauges with record lengths too short to be analysed.csv'. Please check that you do not have the file open.", call. = FALSE)
  }

  #### =========================================== LOOP THROUGH GAUGES =========================================== ####
  # Add progress bar
  pb <- txtProgressBar(min = 0, max = length(gauges), style = 3)

  # Some info
  print(paste(length(gauges), "gauging stations queued for analysis:"))
  writeLines(gauges)
  writeLines(" ")

  myStartTime <- Sys.time()

  PELTresults <- list()

  for (i in 1:length(gauges)){
    print(gauges[i])

    #### ========================================== DATA PRE-PROCESSING ========================================== ####
    preprocessedDataAll <- suppressWarnings(import.preprocess.data(gauges[i], metadata))
    preprocessedData <- preprocessedDataAll$res

    if(is.null(preprocessedDataAll$res)){
      failedInputs[[as.character(gauges[i])]] <- list(gauge = gauges[i], warn = preprocessedDataAll$warn, err = preprocessedDataAll$err)
      next
    }

    inData <- preprocessedData$inData
    gaugeNumber <- preprocessedData$gaugeNumber
    gaugeName <- inData$GaugeName[1]
    rejectedYears <- preprocessedData$rejectedYears

    # Extract just the columns that are required (all years)
    gaugeDataAll <- inData[c("GaugeNumber", "GaugeName", "FirstWaterYear", "LastWaterYear", "Missing", "Rejected", "WaterYear", "Date", "Flow")]
    # Remove rejected years
    gaugeData <- gaugeDataAll[which(!gaugeDataAll$WaterYear %in% rejectedYears),]
    # Remove rows where flow is missing
    gaugeData <- gaugeData[which(!is.na(gaugeData$Flow)),]
    gaugeData$NumberOfYears <- length(gaugeData$WaterYear)  # number actually used (excluding rejected years)

    # Skip gauge if record length is less than 20 years (otherwise it will error as minseglength = 10) and
    # record skipped gauge in file.
    if(gaugeData$NumberOfYears[1] < 20){
      write(c(gaugeNumber, gaugeName, gaugeData$NumberOfYears[1]), "Outputs/PELT/Gauges with record lengths too short to be analysed.csv", ncolumns = 3, sep = ",", append = T)
      next
    }

    #### ======================================= APPLY PELT TEST ======================================= ####
    PELTresults[[i]] <- PELTtest(gaugeData)

    Sys.sleep(0.1)
    setTxtProgressBar(pb, i) # updates progress bar after every gauge has been processed
  }

  # Combine results for each gauge
  PELTdf <- data.frame(do.call(rbind, PELTresults))
  names(PELTdf) <- c("Gauge number", "Gauge name", "Record length (years)", "Number of changes", "Year of change_1", "Year of change_2",
                     "Mean trend direction_1", "Mean trend direction_2",
                     "Mean change_1", "Mean change_2",
                     "Mean pc change_1", "Mean pc change_2",
                     "SD trend direction_1", "SD trend direction_2",
                     "SD change_1", "SD change_2",
                     "SD pc change_1", "SD pc change_2",
                     "Timestamp")

  # Suppress warnings and write out custom error message instead if the file cannot be accessed by R
  writeTest <- suppressWarnings(try(write.csv(PELTdf, "Outputs/PELT/PELT results.csv", row.names = FALSE), silent = TRUE))
  if(class(writeTest) == "try-error"){
    stop("Error saving PELT results file. Please check that you do not have the file open.", call. = FALSE)
  }

  #### ======================================== INPUT DATA WARNINGS AND ERRORS ======================================== ####
  #### ========================================== OUTPUT FILE HEADERS TO CSV ========================================== ####
  # Set up header row
  # Only run this once so don't overwrite
  inputWEHeader <- c("File", "Warnings", "Errors")

  # Suppress warnings and write out custom error message instead if the file cannot be accessed by R
  writeTest <- suppressWarnings(try(write(inputWEHeader, "Outputs/PELT/Input data warnings and errors.csv", ncolumns = length(inputWEHeader), sep = ","), silent = TRUE))
  if(class(writeTest) == "try-error"){
    stop("Error saving input data warnings and errors file. Please check that you do not have the file open.", call. = FALSE)
  }

  #### ========================================= WRITE OUT WARNINGS AND ERRORS ========================================= ####
  # This will only capture the first error as it stops after the first one
  for (file in names(failedInputs)){
    failInfo <- eval(parse(text=paste0("failedInputs$`", as.character(file), "`")))
    Warn <- paste0(failInfo$warn, collapse = " & ")
    Err <- paste0(failInfo$err, collapse = " & ")
    inputWEInfo <- c(failInfo$gauge, Warn, Err)
    # Account for commas in the warning and error messages so that they are not split across columns
    inputWEInfo2 <- paste("\"", inputWEInfo, "\"", sep = "", collapse = ",")
    write(inputWEInfo2, "Outputs/PELT/Input data warnings and errors.csv", ncolumns = length(inputWEHeader), sep = ",", append = T)
  }

  ## Tidy up for loop processing
  close(pb)
  print("*** Total processing time:")
  print(Sys.time()-myStartTime)

}

#### ============================================= PETTITT INTERNAL FUNCTION ============================================= ####
#' PettittTest
#'
#' PettittTest is an internal function used by Pettitt.test to carry out the Pettitt test (using the trend package).
#' @keywords internal
PettittTest <- function(gaugeData, sigLevel){

  # Convert significance level from a percentage to a decimal number
  sigValue <- sigLevel/100

  pettittResults <- list()

  # Implement tests
  pt_test <- pettitt.test(gaugeData$Flow)

  # Find p-value, change point, year of change and interpretation
  p_value <- pt_test$p.value
  cpt <- pt_test$estimate[1]
  year_change <- gaugeData$WaterYear[cpt]
  interpretation <- ifelse(p_value < sigValue, paste0("Reject H0. Statistically significant change in mean at ", sigLevel, "% significance level."),
         paste0("No statistically significant change in mean at ", sigLevel, "% significance level."))

  # Calculate the mean before and after the change and the difference (absolute and percentage)
  mean_change <- mean(gaugeData$Flow[cpt:length(gaugeData$Flow)]) - mean(gaugeData$Flow[1:(cpt-1)])
  mean_percent_change <- (mean_change/mean(gaugeData$Flow[1:(cpt-1)]))*100
  trend_direction <- ifelse(mean_change < 0, "Negative", "Positive")

  # Reset to 1 if p-value is greater than 1
  if(p_value > 1) {p_value = 1}

  # Plot time series segmentation
  plotTitle <- ifelse(nchar(gaugeData$GaugeName[1]) > 0,
                      paste0(gaugeData$GaugeNumber[1], " (", gaugeData$GaugeName[1], ")\nPettitt test"),
                      paste0(gaugeData$GaugeNumber[1], "\nPettitt test"))

  if(p_value > sigValue) {
    g <- ggplot() + geom_line(aes(x = WaterYear, y = Flow, colour = "AMAX"), data = gaugeData, na.rm = T) + theme_bw() + ggtitle(plotTitle) +
      theme(plot.title = element_text(hjust = 0.5)) + ylab(expression(paste("Flow ", (m^3/s)))) + xlab("Water year") +
      scale_colour_manual("", values=c("AMAX" = "black"))
  } else {
    g <- ggplot() + geom_line(aes(x = WaterYear, y = Flow, colour = "AMAX"), data = gaugeData, na.rm = T) + theme_bw() + ggtitle(plotTitle) +
      theme(plot.title = element_text(hjust = 0.5)) +
      geom_vline(data = data.frame(year_change = year_change), aes(xintercept = year_change, colour = "Change point"), na.rm = T, show.legend = FALSE) +
      scale_color_manual("", values = c("AMAX" = "black", "Change point" = "red")) +
      ylab(expression(paste("Flow ", (m^3/s)))) + xlab("Water year")
  }

  outPath <- ifelse(nchar(gaugeData$GaugeName[1]) > 0, paste0("Outputs/Pettitt/", gaugeData$GaugeNumber[1], " (", gaugeData$GaugeName[1], ") Pettitt results.png"),
                    paste0("Outputs/Pettitt/", gaugeData$GaugeNumber[1], " Pettitt results.png"))
  ggsave(filename = outPath, g)

  # Combine results
  pettittResults <- c(gaugeData$GaugeNumber[1], gaugeData$GaugeName[1], p_value, interpretation, year_change,
                   trend_direction, mean_change, mean_percent_change, format(Sys.time(), "%d/%m/%Y %X"))

  return(pettittResults)

}

#### ============================================= PETTITT TEST FUNCTION ============================================= ####
#' Pettitt.test
#'
#' Pettitt.test carries out Pettitt's test, which is designed to detect a sudden change in the mean of a time series.
#' Results are automatically written out to the 'Outputs/Pettitt' folder. The 'Pettitt' folder will be automatically generated
#' if it does not already exist.
#'
#' @param folder string containing a path to a folder containing two sub-folders called 'Inputs' and 'Outputs'. 'Inputs' must contain one file per gauge containing annual maximum data. Data should be .AM files.
#' @param sig.level numeric value giving significance level as a percentage for interpretation of results in the output .csv file (e.g. input ‘5’ to get an interpretation of whether the change point is significant at the 5\% significance level). 5 is the default setting for this parameter.
#'
#' @return A .csv file is automatically generated containing the results for all gauges.
#' @return A plot is also generated for each gauge showing the AMAX data with the detected change point shown as a vertical line, if significant at the specified level.
#' @export
#'
#' @examples
#' # Set the inputs to the function
#' folder <- "C:/Non-stationarity/Data"  # this folder contains sub-folders called 'Inputs' and 'Outputs'
#' sig.level = 5
#' # Run the function
#' Pettitt.test(folder, sig.level)
Pettitt.test <- function(folder, sig.level = 5){
  # Set working directory to be the user-specified folder
  setwd(folder)

  # Check for existence of Input and Output folders and stop and print warning if missing
  if(!file.exists("Inputs") | !file.exists("Outputs")) stop("Inputs and/or Outputs folder does not exist", call. = FALSE)

  # Set up folder for Pettitt outputs if one does not already exist
  if(!file.exists("Outputs/Pettitt")){
    dir.create("Outputs/Pettitt")
  }

  # List the files in the input folder
  gauges <- list.files(paste0(folder, "/Inputs"), pattern = ".AM", full.names = TRUE, ignore.case = TRUE)

  # Import look-up file to extract gauge name to go with the number
  # Suppress warnings and write out custom error message instead if the file cannot be accessed by R
  readTest <- suppressWarnings(try(metadata <- read.csv("Inputs/Gauge_metadata.csv", stringsAsFactors = FALSE), silent = TRUE))
  if(class(readTest) == "try-error"){
    stop("Error reading 'Gauge_metadata.csv' file. Please check that it exists in your 'Inputs' folder.", call. = FALSE)
  }

  # Set up list output for gauges that fail during pre-processing
  failedInputs <- list()

  #### =========================================== LOOP THROUGH GAUGES =========================================== ####
  # Add progress bar
  pb <- txtProgressBar(min = 0, max = length(gauges), style = 3)

  # Some info
  print(paste(length(gauges), "gauging stations queued for analysis:"))
  writeLines(gauges)
  writeLines(" ")

  myStartTime <- Sys.time()

  PettittResultsAll <- list()

  for (i in 1:length(gauges)){
    print(gauges[i])

    #### ========================================== DATA PRE-PROCESSING ========================================== ####
    preprocessedDataAll <- suppressWarnings(import.preprocess.data(gauges[i], metadata))
    preprocessedData <- preprocessedDataAll$res

    if(is.null(preprocessedDataAll$res)){
      failedInputs[[as.character(gauges[i])]] <- list(gauge = gauges[i], warn = preprocessedDataAll$warn, err = preprocessedDataAll$err)
      next
    }

    inData <- preprocessedData$inData
    gaugeNumber <- preprocessedData$gaugeNumber
    rejectedYears <- preprocessedData$rejectedYears

    # Extract just the columns that are required (all years)
    gaugeDataAll <- inData[c("GaugeNumber", "GaugeName", "FirstWaterYear", "LastWaterYear", "Missing", "Rejected", "WaterYear", "Date", "Flow")]
    # Remove rejected years
    gaugeData <- gaugeDataAll[which(!gaugeDataAll$WaterYear %in% rejectedYears),]
    # Remove rows where flow is missing
    gaugeData <- gaugeData[which(!is.na(gaugeData$Flow)),]
    gaugeData$NumberOfYears <- length(gaugeData$WaterYear)  # number actually used (excluding rejected years)

    #### ======================================= APPLY PETTITT TEST ======================================= ####
    PettittResultsAll[[i]] <- PettittTest(gaugeData, sig.level)

    Sys.sleep(0.1)
    setTxtProgressBar(pb, i) # updates progress bar after every gauge has been processed
  }

  # Combine results for each gauge
  PettittDF <- data.frame(do.call(rbind, PettittResultsAll))
  names(PettittDF) <- c("Gauge number", "Gauge name", "p-value","Interpretation", "Year of change", "Trend direction", "Mean change", "Mean pc change", "Timestamp")

  # Suppress warnings and write out custom error message instead if the file cannot be accessed by R
  writeTest <- suppressWarnings(try(write.csv(PettittDF, "Outputs/Pettitt/Pettitt results.csv", row.names = FALSE), silent = TRUE))
  if(class(writeTest) == "try-error"){
    stop("Error saving Pettitt results file. Please check that you do not have the file open.", call. = FALSE)
  }

  #### ======================================== INPUT DATA WARNINGS AND ERRORS ======================================== ####
  #### ========================================== OUTPUT FILE HEADERS TO CSV ========================================== ####
  # Set up header row
  # Only run this once so don't overwrite
  inputWEHeader <- c("File", "Warnings", "Errors")

  # Suppress warnings and write out custom error message instead if the file cannot be accessed by R
  writeTest <- suppressWarnings(try(write(inputWEHeader, "Outputs/Pettitt/Input data warnings and errors.csv", ncolumns = length(inputWEHeader), sep = ","), silent = TRUE))
  if(class(writeTest) == "try-error"){
    stop("Error saving input data warnings and errors file. Please check that you do not have the file open.", call. = FALSE)
  }

  #### ========================================= WRITE OUT WARNINGS AND ERRORS ========================================= ####
  # This will only capture the first error as it stops after the first one
  for (file in names(failedInputs)){
    failInfo <- eval(parse(text=paste0("failedInputs$`", as.character(file), "`")))
    Warn <- paste0(failInfo$warn, collapse = " & ")
    Err <- paste0(failInfo$err, collapse = " & ")
    inputWEInfo <- c(failInfo$gauge, Warn, Err)
    # Account for commas in the warning and error messages so that they are not split across columns
    inputWEInfo2 <- paste("\"", inputWEInfo, "\"", sep = "", collapse = ",")
    write(inputWEInfo2, "Outputs/Pettitt/Input data warnings and errors.csv", ncolumns = length(inputWEHeader), sep = ",", append = T)
  }

  ## Tidy up for loop processing
  close(pb)
  print("*** Total processing time:")
  print(Sys.time()-myStartTime)

}
