#' fit.time.fn.warn.err
#'
#' @param dist distribution type ("GEV" or "GLO")
#' @param gaugeNumber gauge number
#' @param gaugeName gauge name
#' @param gaugeData gauge data
#' @param RPs numeric vector of required return periods
#' @param alpha to set confidence interval in the predict function (1 - CI/100)
#'
#' @return Returns results from fit.time.fn and any warnings or errors associated with running the function.
#' @keywords internal
fit.time.fn.warn.err <- function(dist, gaugeNumber, gaugeName, gaugeData, RPs, alpha){
  warn <- err <- NULL
  res <- withCallingHandlers(
    tryCatch(fit.time.fn(dist, gaugeNumber, gaugeName, gaugeData, RPs, alpha), error=function(e){
      err <<- conditionMessage(e)
      NULL
    }), warning=function(w){
      warn <<- append(warn, conditionMessage(w))
    })
  list(res=res, warn=warn, err=err)
}


#' fit.time.fn
#'
#' Internal function to fit distributions, plot diagnostic plots and calculate return levels.
#'
#' @param dist distribution type ("GEV" or "GLO")
#' @param gaugeNumber gauge number
#' @param gaugeName gauge name
#' @param gaugeData gauge data
#' @param RPs numeric vector of required return periods
#' @param alpha set confidence interval in the predict function (1 - CI/100)
#'
#' @export
#' @keywords internal
fit.time.fn <- function(dist, gaugeNumber, gaugeName, gaugeData, RPs, alpha){

  #### ============================================== EXTREME VALUE ANALYSIS ============================================== ####
  # Fit distributions using the texmex package (MLE with scale parameter on the log scale)
  # Use function set up to catch fitting warnings and errors (in 'Additional_functions.R') - takes a list of the covariates in the location and scale parameters

  distFamily = as.name(tolower(dist))

  # For each type of fit, fit model, find return level estimates and confidence intervals for a range of user-defined return periods and save to a data frame
  for (type in c("stat", "var.loc", "var.scale", "var.loc.scale")){

    if(type == "stat"){
      modelsList <<- list(0, 0)
      Fit = "Stationary"
    } else if(type == "var.loc"){
      modelsList <<- list(~ 1 + ScaledWaterYearIndex, 0)
      Fit = "Non-stationary (varying location)"
    } else if(type == "var.scale"){
      modelsList <<- list(0, ~ 1 + ScaledWaterYearIndex)
      Fit = "Non-stationary (varying scale)"
    } else if(type == "var.loc.scale"){
      modelsList <<- list(~ 1 + ScaledWaterYearIndex, ~ 1 + ScaledWaterYearIndex)
      Fit = "Non-stationary (varying location and scale)"
    }

    # Fit distribution
    set.seed(1000)  # set seed so that bootstrap results are reproducible
    eval(parse(text=paste0("fit.", type, ".all <- suppressWarnings(fitWarning(Flow, gaugeData, distFamily, modelsList, methodType = \"bootstrap\"))")))
    # Extract results
    eval(parse(text=paste0("fit.", type, " <- fit.", type, ".all$res")))

    # Add AIC and BIC
    eval(parse(text=paste0("fit.", type, "$AIC <- 2*length(fit.", type, "$map$par) + 2*-fit.", type, "$map$loglik")))
    eval(parse(text=paste0("fit.", type, "$BIC <- length(fit.", type, "$map$par)*log(length(fit.", type, "$map$data$y)) + 2*-fit.", type, "$map$loglik")))

    # Write out warnings and errors
    if(!is.null(eval(parse(text=paste0("fit.", type, ".all$warn"))))){
      print(paste0("There were warnings when fitting the ", tolower(Fit), " model: ", eval(parse(text=paste0("fit.", type, ".all$warn")))))
    }
    if(!is.null(eval(parse(text=paste0("fit.", type, ".all$err"))))){
      print(paste0("There were errors when fitting the ", tolower(Fit), " model: ", eval(parse(text=paste0("fit.", type, ".all$err")))))
    }

    fitCase <- paste0("fit.", type)

    # Diagnostic plots
    model = eval(parse(text=fitCase))
    outPath <- ifelse(nchar(gaugeName) > 0, paste0("Outputs/Time covariate models/", gaugeNumber, " (", gaugeName, ") ", dist, " diagnostic plots - ", Fit, ".png"),
                      paste0("Outputs/Time covariate models/", gaugeNumber, " ", dist, " diagnostic plots - ", Fit, ".png"))
    eval(parse(text=paste0(fitCase, ".diag.warn.err <- plot.diagnostic.warn.err(dist, model, outPath, gaugeNumber, gaugeName, gaugeData)")))

    # Find estimates and confidence intervals
    eval(parse(text=paste0(type, ".return.levels <- predict(", fitCase, ", M = RPs, ci.fit = TRUE, alpha = alpha, unique.=FALSE)")))

    # Write out results
    # Set up the results table
    resultsTable <- data.frame(WaterYear = gaugeData$WaterYear, AMAX = gaugeData$Flow, WaterYearIndex = gaugeData$WaterYearIndex, ScaledWaterYearIndex = gaugeData$ScaledWaterYearIndex)

    # Add return level results to results table
    # Add stationary estimates first - these results are in a slightly different format from the non-stationary ones
    if(type == "stat"){
      # Add columns of results for every user-defined RP
      dataFrameCols <- paste0("\"", RPs, "-year (lower ", CI, "% CI)\" =
                            signif(unname(stat.return.levels$obj$M.", RPs, "[,3]),4),
                            \"", RPs, "-year estimate\" = signif(unname(stat.return.levels$obj$M.", RPs, "[,2]),4),
                            \"", RPs, "-year (upper ", CI, "% CI)\" = signif(unname(stat.return.levels$obj$M.", RPs, "[,4]),4)",
                              collapse = ",")
      # Add the above into a data frame with the gauge data and the type of fit
      eval(parse(text=paste0("RPestimates.stat <- data.frame(resultsTable, Fit = Fit, ", dataFrameCols, ")")))
    } else{  # else add non-stationary estimates
      # Add columns of results for every user-defined RP
      dataFrameCols <- paste0("\"", RPs, "-year (lower ", CI, "% CI)\" = signif(unname(", type, ".return.levels$obj$M.", RPs, "[,3]),4),
                            \"", RPs, "-year estimate\" = signif(unname(", type, ".return.levels$obj$M.", RPs, "[,2]),4),
                            \"", RPs, "-year (upper ", CI, "% CI)\" = signif(unname(", type, ".return.levels$obj$M.", RPs, "[,4]),4)",
                              collapse = ",")
      # Add the above into a data frame with the type of fit
      eval(parse(text=paste0("RPestimates.", type, " <- data.frame(Fit = Fit, ", dataFrameCols, ")")))
    }
  }

  # Write all errors and warnings from the fitting to a .csv file
  fitWarnErr <- c(gaugeData$GaugeNumber[1], gaugeData$GaugeName[1], paste0(fit.stat.all$warn, collapse = " & "), paste0(fit.stat.all$err, collapse = " & "),
                  paste0(fit.var.loc.all$warn, collapse = " & "), paste0(fit.var.loc.all$err, collapse = " & "),
                  paste0(fit.var.scale.all$warn, collapse = " & "), paste0(fit.var.scale.all$err, collapse = " & "),
                  paste0(fit.var.loc.scale.all$warn, collapse = " & "), paste0(fit.var.loc.scale.all$err, collapse = " & "))
  write(fitWarnErr, paste0("Outputs/Time covariate models/", dist, " model fitting warnings and errors.csv"), ncolumns = length(fitWarnErr), sep = ",", append = T)

  # Compare fits for four models if no errors in diagnostic plots
  if(is.null(fit.stat.diag.warn.err$err) & is.null(fit.var.loc.diag.warn.err$err) & is.null(fit.var.scale.diag.warn.err$err) & is.null(fit.var.loc.scale.diag.warn.err$err)){
    outPathFourModels <- ifelse(nchar(gaugeName) > 0, paste0("Outputs/Time covariate models/", gaugeNumber, " (", gaugeName, ") ", dist, " diagnostic plots - Q-Q plots for all models.png"),
                                paste0("Outputs/Time covariate models/", gaugeNumber, " ", dist, " diagnostic plots - Q-Q plots for all models.png"))
    png(outPathFourModels, width = 1500, height = 1500, res=120)
    comparisonPlotTitle <- ifelse(nchar(gaugeName) > 0,
                                  paste0("Gauging station ", gaugeNumber, " (", gaugeName, ") (", dist, ") \n Q-Q plots for all models (time covariate)"),
                                  paste0("Gauging station ", gaugeNumber, " (", dist, ") \n Q-Q plots for all models (time covariate)"))
    comparisonPlotTitleBold <- textGrob(comparisonPlotTitle, gp=gpar(fontface="bold", fontsize = 20))
    grid.arrange(fit.stat.diag.warn.err$plots$qq + ggtitle(paste0("Stationary: AIC ", round(fit.stat$AIC, digits=2), ", BIC ", round(fit.stat$BIC, digits=2))),
                 fit.var.loc.diag.warn.err$plots$qq+ggtitle(paste0("Location varying: AIC ", round(fit.var.loc$AIC, digits=2), ", BIC ", round(fit.var.loc$BIC, digits=2))),
                 fit.var.scale.diag.warn.err$plots$qq+ggtitle(paste0("Scale varying: AIC ", round(fit.var.scale$AIC, digits=2), ", BIC ", round(fit.var.scale$BIC, digits=2))),
                 fit.var.loc.scale.diag.warn.err$plots$qq+ggtitle(paste0("Location and scale varying: AIC ", round(fit.var.loc.scale$AIC, digits=2), ", BIC ", round(fit.var.loc.scale$BIC, digits=2))),
                 ncol=2, top = comparisonPlotTitleBold)
    dev.off()
  }

  # Save diagnostic plot errors and warnings information for each gauge
  eval(parse(text=paste0("diag.plot.info.", gaugeNumber, " <- list(fit.stat.diag.warn.err = fit.stat.diag.warn.err[1:2], fit.var.loc.diag.warn.err = fit.var.loc.diag.warn.err[1:2], fit.var.scale.diag.warn.err = fit.var.scale.diag.warn.err[1:2], fit.var.loc.scale.diag.warn.err = fit.var.loc.scale.diag.warn.err[1:2])")))

  # Add to list (internal function)
  diagPlotWarningErrorList.int <- list()
  diagPlotWarningErrorList.int[[as.character(gaugeNumber)]] <- eval(parse(text=paste0("diag.plot.info.", gaugeNumber)))

  # Combine the results for each type of fit into one big data frame
  resultsTable <- cbind(RPestimates.stat, RPestimates.var.loc, RPestimates.var.scale, RPestimates.var.loc.scale)

  # Set column names for results table
  # For each user-defined RP create three column names
  namesRP <- paste0("\"", RPs, "-year (lower ", CI, "% CI)\", \"", RPs, "-year\", \"", RPs, "-year (upper ", CI, "% CI)\"", collapse = ",")
  # Repeat four times (one for each fit)
  names <- paste0(rep(paste0("\"Fit\", ", namesRP), 4)[1:4], collapse = ", ")
  # Set all column names incorporating the above
  eval(parse(text=paste0("names(resultsTable) <- c(\"WaterYear\", \"AMAX\", \"WaterYearIndex\", \"ScaledWaterYearIndex\", ", names, ")")))

  # Write out the results table for the gauge
  outPath <- ifelse(nchar(gaugeName) > 0, paste0("Outputs/Time covariate models/", gaugeNumber, " (", gaugeName, ") ", dist, " results.csv"),
                    paste0("Outputs/Time covariate models/", gaugeNumber, " ", dist, " results.csv"))

  # Suppress warnings and write out custom error message instead if the file cannot be accessed by R
  writeTest <- suppressWarnings(try(write.csv(resultsTable, outPath, row.names = FALSE), silent = TRUE))
  if(class(writeTest) == "try-error"){
    stop(paste0("Error saving results file for gauge ", gaugeNumber, ". Please check that you do not have the file open."), call. = FALSE)
  }

  #### ============================================== FIND BEST FITS ============================================== ####

  # Find best fit based on AIC - this is indicative and the user should look at the results and plots as well to make a judgement
  typesAIC <- c(fit.stat$AIC, fit.var.loc$AIC, fit.var.scale$AIC, fit.var.loc.scale$AIC)
  # Set whether each type comes out as best fit
  # Set stationary best fit to 'Yes' if the minimum AIC is from the stationary model. Repeat for non-stationary models.
  # If multiple ones are the best fit then both/all should say 'Yes'.
  bestFitAICStat <- ifelse(min(typesAIC) == fit.stat$AIC, "Yes", "")
  bestFitAICVarLoc <- ifelse(min(typesAIC) == fit.var.loc$AIC, "Yes", "")
  bestFitAICVarScale <- ifelse(min(typesAIC) == fit.var.scale$AIC, "Yes", "")
  bestFitAICVarLocScale <- ifelse(min(typesAIC) == fit.var.loc.scale$AIC, "Yes", "")

  # Find best fit based on BIC - this is indicative and the user should look at the results and plots as well to make a judgement
  typesBIC <- c(fit.stat$BIC, fit.var.loc$BIC, fit.var.scale$BIC, fit.var.loc.scale$BIC)
  # Set whether each type comes out as best fit
  # Set stationary best fit to 'Yes' if the minimum BIC is from the stationary model. Repeat for non-stationary models.
  # If multiple ones are the best fit then both/all should say 'Yes'.
  bestFitBICStat <- ifelse(min(typesBIC) == fit.stat$BIC, "Yes", "")
  bestFitBICVarLoc <- ifelse(min(typesBIC) == fit.var.loc$BIC, "Yes", "")
  bestFitBICVarScale <- ifelse(min(typesBIC) == fit.var.scale$BIC, "Yes", "")
  bestFitBICVarLocScale <- ifelse(min(typesBIC) == fit.var.loc.scale$BIC, "Yes", "")

  # Find best fit based on likelihood ratio testing
  bestModelLabel <- likelihood.ratio.testing(fit.stat, fit.var.loc, fit.var.scale, fit.var.loc.scale)

  # Set whether each type comes out as best fit
  bestFitNllhStat <- ifelse(bestModelLabel == "Stationary", "Yes", "")
  bestFitNllhVarLoc <- ifelse(bestModelLabel == "Non-stationary (varying location)", "Yes", "")
  bestFitNllhVarScale <- ifelse(bestModelLabel == "Non-stationary (varying scale)", "Yes", "")
  bestFitNllhVarLocScale <- ifelse(bestModelLabel == "Non-stationary (varying location and scale)", "Yes", "")

  #### ============================================== WRITE OUT PARAMETERS ============================================== ####
  # Prepare outputs and write parameters, warnings and best fits to summary csv

  # Stationary outputs
  outStationary <- c(gaugeData$GaugeNumber[1], gaugeData$GaugeName[1], gaugeData$NumberOfYears[1], gaugeData$FirstWaterYear[1], gaugeData$LastWaterYear[1],
                        gaugeData$Missing[1], gaugeData$Rejected[1], "Stationary", round(fit.stat$map$par[1],3), round(fit.stat$map$se[1],3),
                        "", "", round(fit.stat$map$par[2],3), round(fit.stat$map$se[2],3), "", "", round(fit.stat$map$par[3],3),
                        round(fit.stat$map$se[3],3), round(fit.stat$AIC,3), round(fit.stat$BIC,3), format(Sys.time(), "%d/%m/%Y %X"),
                        bestFitAICStat, bestFitBICStat, bestFitNllhStat)
  write(outStationary, paste0("Outputs/Time covariate models/", dist, " distribution parameters and fits.csv"), ncolumns = length(outStationary), sep = ",", append = T)

  # Varying location outputs
  outLocVarying <- c(gaugeData$GaugeNumber[1], gaugeData$GaugeName[1], gaugeData$NumberOfYears[1], gaugeData$FirstWaterYear[1], gaugeData$LastWaterYear[1],
                        gaugeData$Missing[1], gaugeData$Rejected[1], "Non-stationary (varying location)", round(fit.var.loc$map$par[1],3),
                        round(fit.var.loc$map$se[1],3), round(fit.var.loc$map$par[2],3), round(fit.var.loc$map$se[2],3),
                        round(fit.var.loc$map$par[3],3), round(fit.var.loc$map$se[3],3), "", "", round(fit.var.loc$map$par[4],3),
                        round(fit.var.loc$map$se[4],3), round(fit.var.loc$AIC,3), round(fit.var.loc$BIC,3),
                        format(Sys.time(), "%d/%m/%Y %X"), bestFitAICVarLoc, bestFitBICVarLoc, bestFitNllhVarLoc)
  write(outLocVarying, paste0("Outputs/Time covariate models/", dist, " distribution parameters and fits.csv"), ncolumns = length(outLocVarying), sep = ",", append = T)

  # Varying scale outputs
  outScaleVarying <- c(gaugeData$GaugeNumber[1], gaugeData$GaugeName[1], gaugeData$NumberOfYears[1], gaugeData$FirstWaterYear[1], gaugeData$LastWaterYear[1],
                          gaugeData$Missing[1], gaugeData$Rejected[1], "Non-stationary (varying scale)", round(fit.var.scale$map$par[1],3),
                          round(fit.var.scale$map$se[1],3), "", "", round(fit.var.scale$map$par[2],3), round(fit.var.scale$map$se[2],3),
                          round(fit.var.scale$map$par[3],3), round(fit.var.scale$map$se[3],3), round(fit.var.scale$map$par[4],3),
                          round(fit.var.scale$map$se[4],3), round(fit.var.scale$AIC,3), round(fit.var.scale$BIC,3),
                          format(Sys.time(), "%d/%m/%Y %X"), bestFitAICVarScale, bestFitBICVarScale, bestFitNllhVarScale)
  write(outScaleVarying, paste0("Outputs/Time covariate models/", dist, " distribution parameters and fits.csv"), ncolumns = length(outScaleVarying), sep = ",", append = T)

  # Varying location and scale outputs
  outLocScaleVarying <- c(gaugeData$GaugeNumber[1], gaugeData$GaugeName[1], gaugeData$NumberOfYears[1], gaugeData$FirstWaterYear[1], gaugeData$LastWaterYear[1],
                             gaugeData$Missing[1], gaugeData$Rejected[1], "Non-stationary (varying location and scale)",
                             round(fit.var.loc.scale$map$par[1],3), round(fit.var.loc.scale$map$se[1],3), round(fit.var.loc.scale$map$par[2],3),
                             round(fit.var.loc.scale$map$se[2],3), round(fit.var.loc.scale$map$par[3],3), round(fit.var.loc.scale$map$se[3],3),
                             round(fit.var.loc.scale$map$par[4],3), round(fit.var.loc.scale$map$se[4],3), round(fit.var.loc.scale$map$par[5],3),
                             round(fit.var.loc.scale$map$se[5],3), round(fit.var.loc.scale$AIC,3), round(fit.var.loc.scale$BIC,3),
                             format(Sys.time(), "%d/%m/%Y %X"), bestFitAICVarLocScale, bestFitBICVarLocScale, bestFitNllhVarLocScale)
  write(outLocScaleVarying, paste0("Outputs/Time covariate models/", dist, " distribution parameters and fits.csv"), ncolumns = length(outLocScaleVarying), sep = ",", append = T)

  # Save fit for each gauge
  eval(parse(text=paste0("fits.dist.fn.int", gaugeNumber, " <- list(dist = dist, data = gaugeData, fit.stat = fit.stat, fit.var.loc = fit.var.loc, fit.var.scale = fit.var.scale, fit.var.loc.scale = fit.var.loc.scale, resultsTable = resultsTable)")))

  # Add to list (internal function)
  fits.int <- list()
  fits.int[[as.character(gaugeNumber)]] <- eval(parse(text=paste0("fits.dist.fn.int", gaugeNumber)))

  return(list(fits.dist.int = fits.int, gaugeNumber = gaugeNumber, gaugeName = gaugeName, diagPlotWarningErrorList.int = diagPlotWarningErrorList.int))
}


#' fit.time
#'
#'fit.time fits stationary and non-stationary GEV or GLO distributions to AMAX flow data for all gauges provided by the user
#'in an ‘Inputs’ folder. This code is based on the 'texmex' package. The three non-stationary models have a varying location parameter, a varying scale parameter and
#'varying both location and scale parameters. The parameters vary with a covariate. This function uses time (an index of
#'water year, centred and scaled) as a covariate and automatically calculates a value for the covariate to associate with each
#'AMAX flow in the input data for the non-stationary distribution fitting, i.e. the user does not need to supply the time covariate.
#'Parameters of the fitted distributions and the results from frequency analysis are automatically written out to .csv files
#'in the 'Outputs/Time covariate models' folder.
#'
#' @param dist string of the distribution to use in the fitting. Must be either "GEV" or "GLO".
#' @param folder string containing a path to a folder containing two sub-folders called 'Inputs' and 'Outputs'. 'Inputs' must contain one file per gauge containing annual maximum data. Data should be .AM files.
#' @param RPs numeric vector of return periods for which frequency analysis is required.
#' @param CI numeric value giving the required confidence interval as a percentage for the results from the flood frequency analysis (e.g. '90' for 90\% is the default value).
#'
#' @return Multiple .csv files are written out: one contains the parameters from the distribution fitting for all gauges, and there is one for each gauge containing the flows from the flood frequency analysis.  The remaining files contain warning and error information.
#' @return A list containing the data used in the analysis, the fitted distributions and the flood frequency results is also returned (but not automatically saved). The information can be accessed by saving the output from running the function to an object in R. The key to the distributions is as follows: stationary fitting = fit.stat, varying location = fit.var.loc, varying scale = fit.var.scale and varying both location and scale = fit.var.loc.scale.
#' @return Diagnostic plots (P-P and Q-Q) are also returned in a range of .png files.
#' @export
#'
#' @examples
#' # Set the inputs to the function
#' dist <- "GEV"
#' folder <- "C:/Non-stationarity/Data"  # this folder contains sub-folders called 'Inputs' and 'Outputs'
#' RPs <- c(2, 10, 20, 50, 100)  # set required return periods to be 2, 10, 20, 50 and 100-year
#' CI <- 90  # set the confidence interval to be 90%
#' # Run the function
#' GEVfits <- fit.time(dist, folder, RPs, CI)  # returns a list containing the data used, the fitted distributions and the frequency analysis results
#' # Save output to a .RDS file if desired
#' saveRDS(GEVfits, "Outputs/Time covariate models/GEV fits from fit time function.RDS")
fit.time <- function(dist, folder, RPs, CI=90){

  alpha <- 1 - CI/100  # to set confidence interval in the predict function

  # Set working directory
  setwd(folder)

  # Set up output list
  fits.dist <- warningErrorList <- failedInputs <- diagPlotWarningErrorList <- list()

  # Check for existence of Input and Output folders and stop and print warning if missing
  if(!file.exists("Inputs") | !file.exists("Outputs")) stop("Inputs and/or Outputs folder does not exist", call. = FALSE)

  # Set up folder for time-only covariate outputs if one does not already exist
  if(!file.exists("Outputs/Time covariate models")){
    dir.create("Outputs/Time covariate models")
  }

  # List the files in the input folder
  gauges <- list.files(paste0(folder, "/Inputs"), pattern = ".AM", full.names = TRUE, ignore.case = TRUE)

  # Import look-up file to extract gauge name to go with the number
  # Suppress warnings and write out custom error message instead if the file cannot be accessed by R
  readTest <- suppressWarnings(try(metadata <- read.csv("Inputs/Gauge_metadata.csv", stringsAsFactors = FALSE), silent = TRUE))
  if(class(readTest) == "try-error"){
    stop("Error reading 'Gauge_metadata.csv' file. Please check that it exists in your 'Inputs' folder.", call. = FALSE)
  }

  #### ========================================== OUTPUT FILE HEADERS TO CSV ========================================== ####
  # Set up header row for parameter output csv
  # Only run this once so don't overwrite
  outHeader <- c("GaugeNumber", "GaugeName", "YearsOfRecord", "FirstWaterYear", "LastWaterYear", "MissingYears", "RejectedYears", "Fit",  "Location/mu0", "Loc/mu0 standard error", "mu1",
                    "mu1 standard error", "ln(scale)/phi0", "ln(scale)/phi0 standard error", "phi1", "phi1 standard error", "Shape", "Shape standard error",
                    "AIC", "BIC", "Time stamp", "BestFitAIC", "BestFitBIC", "BestFitLikelihoodRatioTest")

  # Suppress warnings and write out custom error message instead if the file cannot be accessed by R
  writeTest <- suppressWarnings(try(write(outHeader, paste0("Outputs/Time covariate models/", dist, " distribution parameters and fits.csv"), ncolumns = length(outHeader), sep = ","), silent = TRUE))
  if(class(writeTest) == "try-error"){
    stop("Error saving parameters file. Please check that you do not have the file open.", call. = FALSE)
  }

  # Set up header row for model fitting warnings and errors output csv
  # Only run this once so don't overwrite
  fitWarnErrHeader <- c("Gauge number", "Gauge name", "Stationary model warnings", "Stationary model errors", "Varying location model warnings",
                        "Varying location model errors", "Varying scale model warnings", "Varying scale model errors",
                        "Varying loc and scale model warnings", "Varying loc and scale model errors")

  # Suppress warnings and write out custom error message instead if the file cannot be accessed by R
  writeTest <- suppressWarnings(try(write(fitWarnErrHeader, paste0("Outputs/Time covariate models/", dist, " model fitting warnings and errors.csv"), ncolumns = length(fitWarnErrHeader), sep = ","), silent=TRUE))
  if(class(writeTest) == "try-error"){
    stop("Error saving model fitting warnings and errors file. Please check that you do not have the file open.", call. = FALSE)
  }

  #### ============================================== LOOP THROUGH GAUGES ============================================== ####
  # Add progress bar
  pb <- txtProgressBar(min = 0, max = length(gauges), style = 3)

  # Some info
  print(paste(length(gauges), "gauging stations queued for analysis:"))
  writeLines(gauges)
  writeLines(" ")

  myStartTime <- Sys.time()

  for (i in 1:length(gauges)){

    print(gauges[i])

    #### ============================================== DATA PRE-PROCESSING ============================================== ####
    preprocessedDataAll <- suppressWarnings(import.preprocess.data(gauges[i], metadata))
    preprocessedData <- preprocessedDataAll$res

    if(is.null(preprocessedDataAll$res)){
      failedInputs[[as.character(gauges[i])]] <- list(gauge = gauges[i], warn = preprocessedDataAll$warn, err = preprocessedDataAll$err)
      next
    }

    inData <- preprocessedData$inData
    gaugeNumber <- preprocessedData$gaugeNumber
    gaugeName <- inData$GaugeName[1]
    rejectedYears <- preprocessedData$rejectedYears

    # Keep all years
    gaugeDataAll <- inData[c("GaugeNumber", "GaugeName", "FirstWaterYear", "LastWaterYear", "Missing", "Rejected", "WaterYear", "Date", "Flow")]  # extract just the columns that are required
    # Remove rejected years
    gaugeData <- gaugeDataAll[which(!gaugeDataAll$WaterYear %in% rejectedYears),]
    # Remove rows where flow is missing
    gaugeData <- gaugeData[which(!is.na(gaugeData$Flow)),]
    gaugeData$WaterYearIndex <- gaugeData$WaterYear - gaugeData$WaterYear[1]  # start the water year index column (representing time) at the beginning of the subsetted data
    gaugeData$ScaledWaterYearIndex <- scale(gaugeData$WaterYearIndex) # scales to mean = 0, sd = 1
    gaugeData$NumberOfYears <- length(gaugeData$WaterYear)  # number actually used (excluding rejected years)

    # Run the analysis
    fitRun <- fit.time.fn.warn.err(dist, gaugeNumber, gaugeName, gaugeData, RPs, alpha)

    # Add to list
    fits.dist[[as.character(gaugeNumber)]] <- eval(parse(text=paste0("fitRun$res$fits.dist.int$`", gaugeNumber, "`")))

    # Save warning and error information for each gauge
    # From model fitting
    eval(parse(text=paste0("warn.err.info.", gaugeNumber, " <- list(gaugeName = gaugeName, warning = paste0(fitRun$warn, collapse = \" & \"), error = paste0(fitRun$err, collapse = \" & \"))")))
    # Add to list
    warningErrorList[[as.character(gaugeNumber)]] <- eval(parse(text=paste0("warn.err.info.", gaugeNumber)))
    # From diagnostic plots
    diagPlotWarningErrorList[[as.character(gaugeNumber)]] <- eval(parse(text=paste0("fitRun$res$diagPlotWarningErrorList.int$`",gaugeNumber, "`")))

    Sys.sleep(0.1)
    setTxtProgressBar(pb, i) # updates progress bar after every gauge has been processed
  }

  # Tidy up for loop processing
  close(pb)
  print("*** Total processing time:")
  print(Sys.time()-myStartTime)

  #### ============================================== WARNINGS AND ERRORS ============================================== ####

  #### ================================================== INPUT DATA ================================================== ####
  #### ========================================== OUTPUT FILE HEADERS TO CSV ========================================== ####
  # Set up header row
  # Only run this once so don't overwrite
  inputWEHeader <- c("File", "Warnings", "Errors")

  # Suppress warnings and write out custom error message instead if the file cannot be accessed by R
  writeTest <- suppressWarnings(try(write(inputWEHeader, "Outputs/Time covariate models/Input data warnings and errors.csv", ncolumns = length(inputWEHeader), sep = ","), silent = TRUE))
  if(class(writeTest) == "try-error"){
    stop("Error saving input data warnings and errors file. Please check that you do not have the file open.", call. = FALSE)
  }

  #### ========================================= WRITE OUT WARNINGS AND ERRORS ========================================= ####
  # This will only capture the first error as it stops after the first one
  for (file in names(failedInputs)){
    failInfo <- eval(parse(text=paste0("failedInputs$`", as.character(file), "`")))
    Warn <- paste0(failInfo$warn, collapse = " & ")
    Err <- paste0(failInfo$err, collapse = " & ")
    inputWEInfo <- c(failInfo$gauge, Warn, Err)
    # Account for commas in the warning and error messages so that they are not split across columns
    inputWEInfo2 <- paste("\"", inputWEInfo, "\"", sep = "", collapse = ",")
    write(inputWEInfo2, "Outputs/Time covariate models/Input data warnings and errors.csv", ncolumns = length(inputWEHeader), sep = ",", append = T)
  }

  #### =============================================== FITTING FUNCTION =============================================== ####
  #### ========================================== OUTPUT FILE HEADERS TO CSV ========================================== ####
  # Set up header row
  # Only run this once so don't overwrite
  WEHeader <- c("Gauge", "Gauge Name", "Warnings", "Errors")

  # Suppress warnings and write out custom error message instead if the file cannot be accessed by R
  writeTest <- suppressWarnings(try(write(WEHeader, paste0("Outputs/Time covariate models/", dist, " fit time function warnings and errors.csv"), ncolumns = length(WEHeader), sep = ","), silent = TRUE))
  if(class(writeTest) == "try-error"){
    stop("Error saving fit time function warnings and errors file. Please check that you do not have the file open.", call. = FALSE)
  }

  #### ========================================= WRITE OUT WARNINGS AND ERRORS ========================================= ####
  # This will only capture the first error as it stops after the first one
  for (gauge in names(warningErrorList)){
    gaugeWEs <- eval(parse(text=paste0("warningErrorList$`", as.character(gauge), "`")))
    gaugeName <- gaugeWEs$gaugeName
    Warn <- paste0(gaugeWEs$warning, collapse = " & ")
    Err <- paste0(gaugeWEs$error, collapse = " & ")
    warningErrorInfo <- c(gauge, gaugeName, Warn, Err)
    # Account for commas in the warning and error messages so that they are not split across columns
    warningErrorInfo2 <- paste("\"", warningErrorInfo, "\"", sep = "", collapse = ",")
    write(warningErrorInfo2, paste0("Outputs/Time covariate models/", dist, " fit time function warnings and errors.csv"), ncolumns = length(WEHeader), sep = ",", append = T)
  }

  #### =============================================== DIAGNOSTIC PLOTS =============================================== ####
  #### ========================================== OUTPUT FILE HEADERS TO CSV ========================================== ####
  # Set up header row
  # Only run this once so don't overwrite
  diagWEHeader <- c("Gauge", "Stationary model errors", "Stationary model warnings", "Varying location model errors", "Varying location model warnings",
                "Varying scale model errors", "Varying scale model warnings", "Varying loc and scale model errors", "Varying loc and scale model warnings")

  # Suppress warnings and write out custom error message instead if the file cannot be accessed by R
  writeTest <- suppressWarnings(try(write(diagWEHeader, paste0("Outputs/Time covariate models/", dist, " diagnostic plot warnings and errors.csv"), ncolumns = length(diagWEHeader), sep = ","), silent = TRUE))
  if(class(writeTest) == "try-error"){
    stop("Error saving diagnostic plot warnings and errors file. Please check that you do not have the file open.", call. = FALSE)
  }

  #### ========================================= WRITE OUT WARNINGS AND ERRORS ========================================= ####
  # This will only capture the first error as it stops after the first one
  diagWE <- diagPlotWarningErrorList
  for (gauge in names(diagWE)){
    gaugeWEdata <- eval(parse(text=paste0("diagWE$`", as.character(gauge), "`")))
    statErr <- paste0(gaugeWEdata$fit.stat.diag.warn.err$err, collapse = " & ")
    statWarn <- paste0(gaugeWEdata$fit.stat.diag.warn.err$warn, collapse = " & ")
    varLocErr <- paste0(gaugeWEdata$fit.var.loc.diag.warn.err$err, collapse = " & ")
    varLocWarn <- paste0(gaugeWEdata$fit.var.loc.diag.warn.err$warn, collapse = " & ")
    varScaleErr <- paste0(gaugeWEdata$fit.var.scale.diag.warn.err$err, collapse = " & ")
    varScaleWarn <- paste0(gaugeWEdata$fit.var.scale.diag.warn.err$warn, collapse = " & ")
    varLocScaleErr <- paste0(gaugeWEdata$fit.var.loc.scale.diag.warn.err$err, collapse = " & ")
    varLocScaleWarn <- paste0(gaugeWEdata$fit.var.loc.scale.diag.warn.err$warn, collapse = " & ")
    warningErrorInfo <- c(gauge, statErr, statWarn, varLocErr, varLocWarn, varScaleErr, varScaleWarn, varLocScaleErr, varLocScaleWarn)
    warningErrorInfo2 <- paste("\"", warningErrorInfo, "\"", sep = "", collapse = ",")
    write(warningErrorInfo2, paste0("Outputs/Time covariate models/", dist, " diagnostic plot warnings and errors.csv"), ncolumns = length(diagWEHeader), sep = ",", append = T)
  }

  return(fits.dist)
}
